__all__ = ("gen_all_events",)


import astropy.units as u
import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord, get_sun
from astropy.time import Time

from rubin_scheduler.scheduler.utils import SimTargetooServer, TargetoO
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD, _angular_separation, _hpid2_ra_dec

# Expected start and end for observing run 5
# See: https://observing.docs.ligo.org/plan/
O5_START_MJD = 61557
O5_END_MJD = 62502


def gen_gw_events(mjd_start=O5_START_MJD, mjd_end=O5_END_MJD, scale=1, seed=42):
    """Generate gravitational wave events

    Parameters
    ----------
    mjd_start : `float`
        MJD to start creating events. Default O5_START_MJD set to
        expected O5 run from GW detectors.
    mjd_end : `float`
        MJD to stop creating events. Default O5_END_MJD set to
        expected end of O5 run from GW detectors.
    scale : `float`
        How much to scale the expected number of evenets.
        Expect 22 events if set to default of 1.
    seed : `float`
        Random number generator seed. Default 42.
    """

    """
    Case A) 9 triggers with 30 sq deg area -> should yield 3 visible.
    Case B) 15 triggers with 50 sq deg area -> should yield 5 observable.
    Case C) 24 triggers with 100 sq deg area -> should yield 8 observable.
    Case D) 6 triggers with 250 sq deg area -> should yield 4 observable.
    Case E) 12 triggers with 500 sq deg area -> should yield 2 observable.
    """
    case_2_area = {"A": 30.0, "B": 50, "C": 100, "D": 250, "E": 500}
    case_2_nt = {"A": 9, "B": 15, "C": 24, "D": 6, "E": 12}

    rng = np.random.default_rng(seed=seed)

    n_events = 0
    for case in case_2_nt:
        n_events += case_2_nt[case] * scale

    event_tables = []
    names = ["mjd_start", "ra", "dec", "expires", "radius", "ToO_label"]
    types = [float] * 5 + ["<U50"]

    for case in case_2_area:
        n = case_2_nt[case] * scale
        event_table = np.zeros(int(n), dtype=list(zip(names, types)))
        event_table["radius"] = (case_2_area[case] / np.pi) ** 0.5 * np.pi / 180.0
        event_table["ToO_label"] = "GW_case_%s" % case
        event_tables.append(event_table)

    events = np.concatenate(event_tables)
    n_events = np.size(events)

    # Make sure latitude points spread correctly
    # http://mathworld.wolfram.com/SpherePointPicking.html
    events["ra"] = rng.random(n_events) * np.pi * 2
    events["dec"] = np.arccos(2.0 * rng.random(n_events) - 1.0) - np.pi / 2.0
    events["mjd_start"] = rng.random(n_events) * (mjd_end - mjd_start) + mjd_start
    events["expires"] = events["mjd_start"] + 3

    return events


def gen_bbh_events(mjd_start=O5_START_MJD, mjd_end=O5_END_MJD, scale=1, seed=43):
    """Generate black hole - black home merger events.

    Parameters
    ----------
    mjd_start : `float`
        MJD to start creating events. Default O5_START_MJD set to
        expected O5 run from GW detectors.
    mjd_end : `float`
        MJD to stop creating events. Default O5_END_MJD set to
        expected end of O5 run from GW detectors.
    scale : `float`
        How much to scale the expected number of evenets.
        Expect 21 events if set to default of 1.
    seed : `float`
        Random number generator seed. Default 43.
    """

    rng = np.random.default_rng(seed=seed)
    """
    Case A) 9 triggers with 10 sq deg area -> yields 3 observable
    Case B) 6 triggers with 20 sq deg area -> yields 2 observable
    Case C) 6 triggers with 30 sq deg area -> yields 2 observable
    """
    case_2_area = {"A": 10.0, "B": 20, "C": 30}
    case_2_nt = {"A": 9, "B": 6, "C": 6}

    n_events = 0
    for case in case_2_nt:
        n_events += case_2_nt[case] * scale

    event_tables = []
    names = ["mjd_start", "ra", "dec", "expires", "radius", "ToO_label"]
    types = [float] * 5 + ["<U50"]

    for case in case_2_area:
        n = case_2_nt[case] * scale
        event_table = np.zeros(int(n), dtype=list(zip(names, types)))
        event_table["radius"] = (case_2_area[case] / np.pi) ** 0.5 * np.pi / 180.0
        event_table["ToO_label"] = "BBH_case_%s" % case
        event_tables.append(event_table)

    events = np.concatenate(event_tables)
    n_events = np.size(events)

    # Make sure latitude points spread correctly
    # http://mathworld.wolfram.com/SpherePointPicking.html
    events["ra"] = rng.random(n_events) * np.pi * 2
    events["dec"] = np.arccos(2.0 * rng.random(n_events) - 1.0) - np.pi / 2.0
    events["mjd_start"] = rng.random(n_events) * (mjd_end - mjd_start) + mjd_start
    events["expires"] = events["mjd_start"] + 3

    return events


def gen_neutrino_events(
    scale=1,
    seed=44,
    n_trigger=160,
    mjd_start=SURVEY_START_MJD,
    mjd_end=SURVEY_START_MJD + 3652.5,
    radius=0.0,
):
    """Generate neutrino ToO events

    Parameters
    ----------
    scale : `float`
        How much to scale number of events by. Defualt 1.
    n_trigger : `int`
        Number of events to trigger. Default 160.
    mjd_start : `float`
        Starting MJD. Default SURVEY_START_MJD.
    mjd_end : `float`
        Ending MJD. Default SURVEY_START_MJD + 3652.5.
    radius : `float`
        The search radius of the events on the sky.
        Default 0 (degrees).
    """
    rng = np.random.default_rng(seed=seed)
    n_events = int(n_trigger * scale)

    names = ["mjd_start", "ra", "dec", "expires", "radius", "ToO_label"]
    types = [float] * 5 + ["<U50"]

    events = np.zeros(n_events, dtype=list(zip(names, types)))

    events["ra"] = rng.random(n_events) * np.pi * 2
    events["dec"] = np.arccos(2.0 * rng.random(n_events) - 1.0) - np.pi / 2.0
    events["mjd_start"] = rng.random(n_events) * (mjd_end - mjd_start) + mjd_start
    events["ToO_label"] = "neutrino"
    events["expires"] = events["mjd_start"] + 3
    events["radius"] = np.radians(radius)

    return events


def gen_lensed_BNS(mjd_start=O5_START_MJD, mjd_end=O5_END_MJD, scale=1, seed=45):
    """Generate lensed black hole - neutron star GW events.

    Parameters
    ----------
    scale : `float`
        How much to scale number of events by. Default 1.
        Expect 15 events with default scale.
    mjd_start : `float`
        Starting MJD. Default O5_START_MJD
    mjd_end : `float`
        Ending MJD. Defualt O5_END_MJD
    """

    rng = np.random.default_rng(seed=seed)

    """
    Case A) 9 triggers with 10 sq deg area -> yields 3 observable
    Case B) 6 triggers with 20 sq deg area -> yields 2 observable
    """

    case_2_area = {"A": 900, "B": 15}
    case_2_nt = {"A": 1, "B": 1}

    n_events = 0
    for case in case_2_nt:
        n_events += case_2_nt[case] * scale

    event_tables = []
    names = ["mjd_start", "ra", "dec", "expires", "radius", "ToO_label"]
    types = [float] * 5 + ["<U50"]

    for case in case_2_area:
        n = case_2_nt[case] * scale
        event_table = np.zeros(int(n), dtype=list(zip(names, types)))
        event_table["radius"] = (case_2_area[case] / np.pi) ** 0.5 * np.pi / 180.0
        event_table["ToO_label"] = "lensed_BNS_case_%s" % case
        event_tables.append(event_table)

    events = np.concatenate(event_tables)
    n_events = np.size(events)

    # Make sure latitude points spread correctly
    # http://mathworld.wolfram.com/SpherePointPicking.html
    events["ra"] = rng.random(n_events) * np.pi * 2
    events["dec"] = np.arccos(2.0 * rng.random(n_events) - 1.0) - np.pi / 2.0
    events["mjd_start"] = rng.random(n_events) * (mjd_end - mjd_start) + mjd_start
    events["expires"] = events["mjd_start"] + 3

    return events


def gen_sso_events(n_events=900, twi_fraction=0.75, seed=52, radius=2.0, mjd_start=SURVEY_START_MJD):
    """Generate solar system ToO events

    Originally set to 300 events, swapped to a more approximate
    generation of events and increased number of events to 900 to
    ensure enough trigger.

    Parameters
    ----------
    n_events : `int`
        The number of events to generate. Default 900
    twi_fraction : `float`
        Fraction of events to set as earth-interior alerts,
        e.g., events that are near the sun and only visible at
        twilight. Default 0.75.
    seed : `float`
        Random number seed. Default 52.
    radius : `float`
        Search radius on sky for events. Default 2 (degrees).
    mjd_start : `float`
        MJD of when to start generating events.
        Default SURVEY_START_MJD.
    """

    # Seems like solar elongations are around 43.3 degrees
    # with 2.05 degree standard deviation
    elong_mean = np.radians(43.3)
    elong_std = np.radians(2.05)
    pad = np.radians(0.5)

    # Then ecliptic latitude between -40 and +15

    rng = np.random.default_rng(seed=seed)

    nside = 128
    ra_rad, dec_rad = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
    coord = SkyCoord(ra=ra_rad * u.rad, dec=dec_rad * u.rad, frame="icrs")
    eclip_lat_rad = coord.barycentrictrueecliptic.lat.rad

    near_eclip = np.where((np.radians(-40.0) < eclip_lat_rad) & (eclip_lat_rad < np.radians(15.0)))[0]

    eclip_mask = np.zeros(ra_rad.size, dtype=bool)
    eclip_mask[near_eclip] = True

    # Generate inner SS objects
    names = ["mjd_start", "ra", "dec", "expires", "radius", "ToO_label"]
    types = [float] * 5 + ["<U50"]
    # Array to hold events
    twi_events = np.zeros(int(n_events * twi_fraction), dtype=list(zip(names, types)))

    n_inner = int(n_events * twi_fraction)
    # We want to trigger on nights divisible by 4+1
    potential_mjds = np.arange(0, 3652, 4) + mjd_start + 1
    twi_events["mjd_start"] = rng.choice(potential_mjds, size=n_inner, replace=False)

    times = Time(twi_events["mjd_start"], format="mjd")
    sun = get_sun(times)

    ras = []
    decs = []

    for i, time in enumerate(times):
        sun_dist = _angular_separation(sun[i].ra.rad, sun[i].dec.rad, ra_rad, dec_rad)
        elong_draw = rng.normal(loc=elong_mean, scale=elong_std, size=1)
        good = np.where(((elong_draw - pad) < sun_dist) & (sun_dist < (elong_draw + pad)) & eclip_mask)[0]

        ra, dec = _hpid2_ra_dec(nside, rng.choice(good, size=1))
        ras.append(ra[0])
        decs.append(dec[0])

    twi_events["ra"] = ras
    twi_events["dec"] = decs

    twi_events["ToO_label"] = "SSO_twilight"

    reg_events = np.zeros(int(n_events - twi_events.size), dtype=list(zip(names, types)))

    potential_mjds = np.arange(0, 3652) + mjd_start
    reg_events["mjd_start"] = rng.choice(potential_mjds, size=reg_events.size, replace=False)
    times = Time(reg_events["mjd_start"], format="mjd")
    sun = get_sun(times)

    ras = []
    decs = []

    for i, time in enumerate(times):
        sun_dist = _angular_separation(sun[i].ra.rad, sun[i].dec.rad, ra_rad, dec_rad)
        good = np.where(
            (sun_dist < np.radians(160.0))
            & (sun_dist > np.radians(90.0))
            & (eclip_lat_rad > np.radians(-80))
            & (eclip_lat_rad < np.radians(25))
        )[0]
        ra, dec = _hpid2_ra_dec(nside, rng.choice(good, size=1))
        ras.append(ra[0])
        decs.append(dec[0])

    reg_events["ra"] = ras
    reg_events["dec"] = decs

    reg_events["ToO_label"] = "SSO_night"

    events = np.concatenate([twi_events, reg_events])
    events["expires"] = events["mjd_start"] + 3
    events["radius"] = np.radians(radius)

    return events


def gen_all_events(scale=1, nside=DEFAULT_NSIDE, include_gw=True, include_neutrino=True, include_ss=True):
    """Function to generate ToO events

    Parameters
    ----------
    scale : `float`
        Amount to scale the total number of events. Default 1.
    nside : `int`
        HEALpix nside. Default set to DEAFAULT_NSIDE.
    include_gw : `bool`
        Include gravitational wave events (GW, BH-BH, and lensed BH-NS events).
        Default True.
    include_neutrino : `bool`
        Include neutrino events. Default True.
    include_ss : `bool`
        Include solar system events. Default True.
    """

    if scale == 0:
        return None, None

    ra, dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))

    events = []
    if include_gw:
        events.append(gen_gw_events(scale=3.0 * scale))
        events.append(gen_bbh_events(scale=3.0 * scale))
        events.append(gen_lensed_BNS(scale=3.0 * scale))
    # Not varying the number of neutrino events
    if include_neutrino:
        events.append(gen_neutrino_events())
    # Not varying the number of SSO events
    if include_ss:
        events.append(gen_sso_events())
    event_table = np.concatenate(events)

    event_table.sort(order="mjd_start")

    events = []
    for i, event_time in enumerate(event_table["mjd_start"]):
        dist = _angular_separation(ra, dec, event_table["ra"][i], event_table["dec"][i])
        good = np.where(dist <= event_table["radius"][i])
        footprint = np.zeros(ra.size, dtype=float)
        footprint[good] = 1
        events.append(
            TargetoO(
                i,
                footprint,
                event_time,
                event_table["expires"][i] - event_table["mjd_start"][i],
                ra_rad_center=event_table["ra"][i],
                dec_rad_center=event_table["dec"][i],
                too_type=event_table["ToO_label"][i],
            )
        )
    events = SimTargetooServer(events)

    return events, event_table



__all__ = (
    "get_ideal_model_observatory",
    "update_model_observatory_sunset",
    "standard_masks",
    "simple_rewards",
    "simple_pairs_survey",
    "simple_greedy_survey",
    "simple_rewards_field_survey",
    "simple_field_survey",
)


import numpy as np
from astropy.time import Time

import rubin_scheduler.scheduler.basis_functions as basis_functions
import rubin_scheduler.scheduler.detailers as detailers
import rubin_scheduler.scheduler.features as features
from rubin_scheduler.scheduler.model_observatory import (
    KinemModel,
    ModelObservatory,
    rotator_movement,
    tma_movement,
)
from rubin_scheduler.scheduler.schedulers import BandSwapScheduler
from rubin_scheduler.scheduler.surveys import BlobSurvey, FieldSurvey, GreedySurvey
from rubin_scheduler.scheduler.utils import Footprint, get_current_footprint
from rubin_scheduler.site_models import Almanac, ConstantSeeingData, ConstantWindData
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD


def get_ideal_model_observatory(
    nside: int = DEFAULT_NSIDE,
    dayobs: str = "2024-09-09",
    fwhm_500: float = 1.6,
    wind_speed: float = 5.0,
    wind_direction: float = 340,
    tma_percent: float = 70,
    rotator_percent: float = 100,
    survey_start: Time = Time("2024-09-09T12:00:00", format="isot", scale="utc").mjd,
) -> ModelObservatory:
    """Set up a model observatory with constant seeing and wind speed,
    running on 'dayobs'.

    Track a constant 'survey_start' over different 'dayobs' so that the
    'night' value in the simulation outputs is consistent, and any
    basis functions which might use 'season' values from the conditions
    have appropriate values.

    Parameters
    ----------
    nside : `int`
        The nside for the model observatory.
    dayobs : `str`
        DAYOBS formatted str (YYYY-MM-DD) for the evening to start
        up the observatory.
    fwhm_500 : `float`, optional
        The value to set for atmospheric component of seeing,
         constant seeing throughout the night (arcseconds).
        Ad-hoc value for start of comcam on-sky operations about 2.0".
    wind_speed : `float`, optional
        Set a (constant) wind speed for the night, (m/s).
        Default of 5.0 is minimal but noticeable.
    wind_direction : `float`, optional
        Set a (constant) wind direction for the night (deg).
        Default of 340 is midrange between typical wind directions for
        the site (see SITCOMTN-126).
    tma_percent : `float`, optional
        Set a percent of full-performance for the telescope TMA (0-100).
        Value of 10(%) is likely for start of comcam on-sky SV surveys.
    rotator_percent : `float`, optional
        Set a percent of full-performance for the rotator.
        Default of 100% is likely for the start of comcam on-sky SV surveys.
    survey_start : `float`, optional
        MJD of the day of the survey start of operations.
        This should be kept constant for a given set of simulations,
        so that the "night" value in the output is consistent.
        For surveys which use basis functions depending on season, this
        is also important to be constant.

    Returns
    -------
    observatory : `~.scheduler.model_observatory.ModelObservatory`
        A ModelObservatory set up to start operations in the evening
        of DAYOBS.

    Notes
    -----
    The time for the model observatory will be advanced to the time
    of `sunset_start_key` (default -12 degree sunset) in the model
    observatory. The bands may not be correct however; use
    `update_model_observatory_sunset` to get bands in place.
    """
    # Set up a fresh model observatory
    mjd_now = Time(f"{dayobs}T12:00:00", format="isot", scale="utc").mjd

    kinematic_model = KinemModel(mjd0=mjd_now)
    rot = rotator_movement(rotator_percent)
    kinematic_model.setup_camera(**rot)
    tma = tma_movement(tma_percent)
    kinematic_model.setup_telescope(**tma)

    # Some weather telemetry that might be useful
    seeing_data = ConstantSeeingData(fwhm_500=fwhm_500)
    wind_data = ConstantWindData(wind_direction=wind_direction, wind_speed=wind_speed)

    # Set up the model observatory
    observatory = ModelObservatory(
        nside=nside,
        mjd=mjd_now,
        mjd_start=survey_start,
        kinem_model=kinematic_model,  # Modified kinematics
        cloud_data="ideal",  # No clouds
        seeing_data=seeing_data,  # Modified seeing
        wind_data=wind_data,  # Add some wind
        downtimes="ideal",  # No downtime
        lax_dome=True,  # dome crawl?
        init_load_length=1,  # size of skybrightness files to load first
    )
    return observatory


def update_model_observatory_sunset(
    observatory: ModelObservatory, band_scheduler: BandSwapScheduler, twilight: int | float = -12
) -> ModelObservatory:
    """Ensure correct bands are in place according to the band_scheduler.

    Parameters
    ----------
    observatory : `~.scheduler.model_observatory.ModelObservatory`
        The ModelObservatory simulating the observatory.
    band_scheduler : `~.scheduler.schedulers.BandScheduler`
        The band scheduler providing appropriate information on
        the bands that should be in place on the current observatory day.
    twilight : `int` or `float`
        If twilight is -12 or -18, the Almanac -12 or -18 degree twilight
        times are used to set the current observatory time.
        If  any other value is provided, it is assumed to be a specific
        MJD to start operating the observatory.
        Band choices are based on the time after advancing to twilight.

    Returns
    -------
    observatory :  `~.scheduler.model_observatory.ModelObservatory`
        The ModelObservatory simulating the observatory, updated to the
        time of 'twilight' and with mounted_bands matching
        the bands chosen by the band_scheduler for the current time
        at 'twilight'.
    """
    # Move to *next* possible sunset
    if twilight in (-12, -18):
        time_key = f"sun_n{twilight*-1 :0d}_setting"
        # Use the observatory almanac to find the exact time to forward to
        alm_indx = np.searchsorted(observatory.almanac.sunsets[time_key], observatory.mjd, side="right") - 1
        new_mjd = observatory.almanac.sunsets[time_key][alm_indx]
        if new_mjd < observatory.mjd:
            alm_indx += 1
            new_mjd = observatory.almanac.sunsets[time_key][alm_indx]
        observatory.mjd = new_mjd
    else:
        # Assume twilight was a particular (MJD) time
        observatory.mjd = twilight

    # Make sure correct bands are mounted
    conditions = observatory.return_conditions()
    bands_needed = band_scheduler(conditions)
    observatory.observatory.mount_bands(bands_needed)
    return observatory


def standard_masks(
    nside: int,
    moon_distance: float = 30.0,
    wind_speed_maximum: float = 20.0,
    min_alt: float = 20,
    max_alt: float = 86.5,
    min_az: float = 0,
    max_az: float = 360,
    shadow_minutes: float = 30,
) -> list[basis_functions.BaseBasisFunction]:
    """A set of standard mask functions.

    Avoids the moon, bright planets, high wind, and
    areas on the sky out of bounds, using
    the MoonAvoidanceBasisFunction, PlanetMaskBasisFunction,
    AvoidDirectWindBasisFunction, and the AltAzShadowMaskBasisFunction.

    Parameters
    ----------
    nside : `int` or None
        The healpix nside to use.
        Default of None uses rubin_scheduler.utils.get_default_nside.
    moon_distance : `float`, optional
        Moon avoidance distance, in degrees.
    wind_speed_maximum : `float`, optional
        Wind speed maximum to apply to the wind avoidance basis function,
        in m/s.
    min_alt : `float`, optional
        Minimum altitude (in degrees) to observe.
    max_alt : `float`, optional
        Maximum altitude (in degrees) to observe.
    min_az : `float`, optional
        Minimum azimuth angle (in degrees) to observe.
    max_az : `float`, optional
        Maximum azimuth angle (in degrees) to observe.
    shadow_minutes : `float`, optional
        Avoid inaccessible alt/az regions, as well as parts of the sky
        which will move into those regions within `shadow_minutes` (minutes).

    Returns
    -------
    mask_basis_functions : `list` [`BaseBasisFunction`]
        Mask basis functions should always be used with a weight of 0.
        The masked (np.nan or -np.inf) regions will remain masked,
        but the basis function values won't influence the reward.
    """
    masks = []
    # Add the Moon avoidance mask
    masks.append(basis_functions.MoonAvoidanceBasisFunction(nside=nside, moon_distance=moon_distance))
    # Add a mask around bright planets
    masks.append(basis_functions.PlanetMaskBasisFunction(nside=nside))
    # Add the wind avoidance mask
    masks.append(basis_functions.AvoidDirectWind(nside=nside, wind_speed_maximum=wind_speed_maximum))
    # Avoid inaccessible parts of the sky, as well as places that will
    # move into those places within shadow_minutes.
    masks.append(
        basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside,
            min_alt=min_alt,
            max_alt=max_alt,
            min_az=min_az,
            max_az=max_az,
            shadow_minutes=shadow_minutes,
        )
    )
    return masks


def simple_rewards(
    footprints: Footprint,
    bandname: str,
    nside: int = DEFAULT_NSIDE,
    m5_weight: float = 6.0,
    footprint_weight: float = 1.5,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    repeat_weight: float = -20,
) -> list[basis_functions.BaseBasisFunction]:
    """A simple set of rewards for area-based surveys.

    Parameters
    ----------
    footprints : `Footprint`
        A Footprint class, which takes a target map and adds a
        time-dependent weighting on top (such as seasonal weighting).
    bandname : `str`
        The bandname active for these rewards.
    nside : `int`, optional
        The nside for the rewards.
    m5_weight : `float`, optional
        The weight to give the M5Diff basis function.
    footprint_weight : `float`, optional
        The weight to give the footprint basis function.
    slewtime_weight : `float`, optional
        The weight to give the slewtime basis function.
    stayband_weight : `float`, optional
        The weight to give the BandChange basis function.
    repeat_weight : `float`, optional
        The weight to give the VisitRepeat basis function.
        This is negative by default, to avoid revisiting the same part
        of the sky within `gap_max` (3 hours) in any band (except in the
        pairs survey itself).

    Returns
    -------
    reward_functions : `list` [(`BaseBasisFunction`, `float`)]
        List of tuples, each tuple is a reward function followed by
        its respective weight.

    Notes
    -----
    Increasing the m5_weight moves visits toward darker skies.
    Increasing the footprint weight distributes visits more evenly.
    Increasing the slewtime weight acquires visits with shorter slewtime.
    The balance in the defaults here has worked reasonably for pair
    surveys in simulations.
    """
    reward_functions = []
    # Add M5 basis function (rewards dark sky)
    reward_functions.append((basis_functions.M5DiffBasisFunction(bandname=bandname, nside=nside), m5_weight))
    # Add a footprint basis function
    # (rewards sky with fewer visits)
    reward_functions.append(
        (
            basis_functions.FootprintBasisFunction(
                bandname=bandname,
                footprint=footprints,
                out_of_bounds_val=np.nan,
                nside=nside,
            ),
            footprint_weight,
        )
    )
    # Add a reward function for small slewtimes.
    reward_functions.append(
        (
            basis_functions.SlewtimeBasisFunction(bandname=bandname, nside=nside),
            slewtime_weight,
        )
    )
    # Add a reward to stay in the same band as much as possible.
    reward_functions.append((basis_functions.BandChangeBasisFunction(bandname=bandname), stayband_weight))
    # And this is technically a mask, to avoid asking for observations
    # which are not possible. However, it depends on the bands
    # requested, compared to the bands available in the camera.
    reward_functions.append((basis_functions.BandLoadedBasisFunction(bandnames=bandname), 0))
    # And add a basis function to avoid repeating the same pointing
    # (VisitRepeat can be used to either encourage or discourage repeats,
    # depending on the repeat_weight value).
    reward_functions.append(
        (
            basis_functions.VisitRepeatBasisFunction(
                gap_min=0, gap_max=3 * 60.0, bandname=None, nside=nside, npairs=20
            ),
            repeat_weight,
        )
    )
    return reward_functions


def simple_pairs_survey(
    nside: int = DEFAULT_NSIDE,
    bandname: str = "g",
    bandname2: str | None = None,
    mask_basis_functions: list[basis_functions.BaseBasisFunction] | None = None,
    reward_basis_functions: list[basis_functions.BaseBasisFunction] | None = None,
    reward_basis_functions_weights: list[float] | None = None,
    survey_start: float = SURVEY_START_MJD,
    footprints_hp: np.ndarray | None = None,
    footprint: Footprint | None = None,
    camera_rot_limits: list[float] = [-80.0, 80.0],
    pair_time: float = 30.0,
    exptime: float = 30.0,
    nexp: int = 1,
    science_program: str | None = None,
    observation_reason: str | None = None,
    dither: str = "night",
    camera_dither: str = "night",
    require_time: bool = False,
) -> BlobSurvey:
    """Set up a simple blob survey to acquire pairs of visits.

    Parameters
    ----------
    nside  : `int`, optional
        Nside for the surveys.
    bandname : `str`, optional
        Bandname for the first visit of the pair.
    bandname2 : `str` or None, optional
        Bandname for the second visit of the pair. If None, the
        first band will be used for both visits.
    mask_basis_functions : `list` [`BaseBasisFunction`] or None
        List of basis functions to use as masks (with implied weight 0).
        If None, `standard_masks` is used with default parameters.
    reward_basis_functions : `list` [`BaseBasisFunction`] or None
        List of basis functions to use as rewards.
        If None, a basic set of basis functions will be used.
    reward_basis_functions_weights : `list` [`float`] or None
        List of values to use as weights for the reward basis functions.
        If None, default values for the basic set will be used.
    survey_start : `float` or None
        The start of the survey, in MJD.
        If None, `survey_start_mjd()` is used.
        This should be the start of the survey, not the current time.
    footprints_hp : `np.ndarray` (N,) or None
        An array of healpix maps with the target survey area, with dtype
        like [(bandname, '<f8'), (bandname2, '<f8')].
        If None, `get_current_footprint()` will be used, which will cover
        the expected LSST survey footprint.
    camera_rot_limits : `list` [`float`]
        The rotator limits to expect for the camera.
        These should be slightly padded from true limits, to allow for
        slight delays between requesting observations and acquiring them.
    pair_time : `float`
        The ideal time between pairs of visits, in minutes.
    exptime : `float`
        The on-sky exposure time per visit.
    nexp : `int`
        The number of exposures per visit (exptime * nexp = total on-sky time).
    science_program : `str` | None
        The science_program for the Survey.
        This maps to the BLOCK and `science_program` in the consDB.
    observation_reason : `str` | None
        The observation_reason for the Survey.
        Indicates survey mode, potentially enhanced with particular reason.
    require_time : `bool`
        If True, add a mask basis function that checks there is enough
        time before twilight to execute the pairs. Default False.

    Returns
    -------
    pair_survey : `BlobSurvey`
        A blob survey configured to take pairs at spacing of pair_time,
        in bandname + bandname2.
    """

    # Use the Almanac to find the position of the sun at the start of survey.
    almanac = Almanac(mjd_start=survey_start)
    sun_moon_info = almanac.get_sun_moon_positions(survey_start)
    sun_ra_start = sun_moon_info["sun_RA"].copy()

    if footprint is None:
        if footprints_hp is None:
            footprints_hp, labels = get_current_footprint(nside=nside)
        footprints = Footprint(mjd_start=survey_start, sun_ra_start=sun_ra_start, nside=nside)
        for f in footprints_hp.dtype.names:
            footprints.set_footprint(f, footprints_hp[f])
    else:
        footprints = footprint

    if mask_basis_functions is None:
        mask_basis_functions = standard_masks(nside=nside)

    # Don't start a blob if there isn't time to finish it before twilight
    if require_time:
        time_needed = pair_time
        if bandname2 is not None:
            time_needed *= 2
        mask_basis_functions.append(basis_functions.TimeToTwilightBasisFunction(time_needed=time_needed))

    # Mask basis functions have zero weights.
    mask_basis_functions_weights = [0 for mask in mask_basis_functions]

    if reward_basis_functions is None:
        # Create list of (tuples of) basic reward basis functions and weights.
        m5_weight = 6.0
        footprint_weight = 1.5
        slewtime_weight = 3.0
        stayband_weight = 3.0
        repeat_weight = -20

        if bandname2 is None:
            reward_functions = simple_rewards(
                footprints=footprints,
                bandname=bandname,
                nside=nside,
                m5_weight=m5_weight,
                footprint_weight=footprint_weight,
                slewtime_weight=slewtime_weight,
                stayband_weight=stayband_weight,
                repeat_weight=repeat_weight,
            )

        else:
            # Add the same basis functions, but M5 and footprint
            # basis functions need to be added twice, with half the weight.
            rf1 = simple_rewards(
                footprints=footprints,
                bandname=bandname,
                nside=nside,
                m5_weight=m5_weight / 2.0,
                footprint_weight=footprint_weight / 2.0,
                slewtime_weight=slewtime_weight,
                stayband_weight=stayband_weight,
                repeat_weight=repeat_weight,
            )
            rf2 = simple_rewards(
                footprints=footprints,
                bandname=bandname2,
                nside=nside,
                m5_weight=m5_weight / 2.0,
                footprint_weight=footprint_weight / 2.0,
                slewtime_weight=0,
                stayband_weight=0,
                repeat_weight=0,
            )
            # Now clean up and combine these - and remove the separate
            # BasisFunction for BandLoadedBasisFunction.
            reward_functions = [(i[0], i[1]) for i in rf1 if i[1] > 0] + [
                (i[0], i[1]) for i in rf2 if i[1] > 0
            ]
            # Then put back in the BandLoadedBasisFunction with both bands.
            bandnames = [fn for fn in [bandname, bandname2] if fn is not None]
            reward_functions.append((basis_functions.BandLoadedBasisFunction(bandnames=bandnames), 0))

        # unpack the basis functions and weights
        reward_basis_functions_weights = [val[1] for val in reward_functions]
        reward_basis_functions = [val[0] for val in reward_functions]

    # Set up blob surveys.
    if bandname2 is None:
        survey_name = "simple pair %i, %s" % (pair_time, bandname)
    else:
        survey_name = "simple pair %i, %s%s" % (pair_time, bandname, bandname2)

    # Set up detailers for each requested observation.
    detailer_list = []
    # Avoid camera rotator limits.
    detailer_list.append(
        detailers.CameraRotDetailer(
            min_rot=np.min(camera_rot_limits), max_rot=np.max(camera_rot_limits), dither=camera_dither
        )
    )
    # Convert rotTelPos to rotSkyPos_desired
    detailer_list.append(detailers.Rottep2RotspDesiredDetailer(telescope="rubin"))
    # Reorder visits in a blob so that closest to current altitude is first.
    detailer_list.append(detailers.CloseAltDetailer())
    # Sets a flush-by date to avoid running into prescheduled visits.
    detailer_list.append(detailers.FlushForSchedDetailer())
    # Add a detailer to label visits as either first or second of the pair.
    if bandname2 is not None:
        detailer_list.append(detailers.TakeAsPairsDetailer(bandname=bandname2))

    # Set up the survey.
    ignore_obs = ["DD"]

    BlobSurvey_params = {
        "slew_approx": 7.5,
        "band_change_approx": 140.0,
        "read_approx": 2.4,
        "flush_time": pair_time * 3,
        "smoothing_kernel": None,
        "nside": nside,
        "seed": 42,
        "dither": dither,
        "twilight_scale": False,
    }

    pair_survey = BlobSurvey(
        reward_basis_functions + mask_basis_functions,
        reward_basis_functions_weights + mask_basis_functions_weights,
        bandname1=bandname,
        bandname2=bandname2,
        exptime=exptime,
        ideal_pair_time=pair_time,
        survey_name=survey_name,
        observation_reason=observation_reason,
        ignore_obs=ignore_obs,
        nexp=nexp,
        detailers=detailer_list,
        science_program=science_program,
        **BlobSurvey_params,
    )

    # Tucking this here so we can look at how many observations
    # recorded for this survey and what was the last one.
    pair_survey.extra_features["ObsRecorded"] = features.NObsCount()
    pair_survey.extra_features["LastObs"] = features.LastObservation()

    return pair_survey


def simple_greedy_survey(
    nside: int = DEFAULT_NSIDE,
    bandname: str = "r",
    mask_basis_functions: list[basis_functions.BaseBasisFunction] | None = None,
    reward_basis_functions: list[basis_functions.BaseBasisFunction] | None = None,
    reward_basis_functions_weights: list[float] | None = None,
    survey_start: float = SURVEY_START_MJD,
    footprints_hp: np.ndarray | None = None,
    camera_rot_limits: list[float] = [-80.0, 80.0],
    exptime: float = 30.0,
    nexp: int = 1,
    science_program: str | None = None,
    observation_reason: str | None = None,
    dither: str = "night",
    camera_dither: str = "night",
) -> GreedySurvey:
    """Set up a simple greedy survey to just observe single visits.

    Parameters
    ----------
    nside  : `int`, optional
        Nside for the surveys.
    bandname : `str`, optional
        Bandname for the visits.
    mask_basis_functions : `list` [`BaseBasisFunction`] or None
        List of basis functions to use as masks (with implied weight 0).
        If None, `standard_masks` is used with default parameters.
    reward_basis_functions : `list` [`BaseBasisFunction`] or None
        List of basis functions to use as rewards.
        If None, a basic set of basis functions will be used.
    reward_basis_functions_weights : `list` [`float`] or None
        List of values to use as weights for the reward basis functions.
        If None, default values for the basic set will be used.
    survey_start : `float` or None
        The start of the survey, in MJD.
        If None, `survey_start_mjd()` is used.
        This should be the start of the survey, not the current time.
    footprints_hp : `np.ndarray` (N,) or None
        An array of healpix maps with the target survey area, with dtype
        like [(bandname, '<f8'), (bandname2, '<f8')].
        If None, `get_current_footprint()` will be used, which will cover
        the expected LSST survey footprint.
    camera_rot_limits : `list` [`float`]
        The rotator limits to expect for the camera.
        These should be slightly padded from true limits, to allow for
        slight delays between requesting observations and acquiring them.
    exptime : `float`
        The on-sky exposure time per visit.
    nexp : `int`
        The number of exposures per visit (exptime * nexp = total on-sky time).
    science_program : `str` | None
        The science_program key for the FieldSurvey.
        This maps to the BLOCK and `science_program` in the consDB.
    observation_reason : `str` | None
        The observation_reason for the Survey.
        Indicates survey mode, potentially enhanced with particular reason.

    Returns
    -------
    greedy_survey : `GreedySurvey`
        A greedy survey configured to take the next best (single) visit
        in bandname.
    """

    # Use the Almanac to find the position of the sun at the start of survey.
    almanac = Almanac(mjd_start=survey_start)
    sun_moon_info = almanac.get_sun_moon_positions(survey_start)
    sun_ra_start = sun_moon_info["sun_RA"].copy()

    if footprints_hp is None:
        footprints_hp, labels = get_current_footprint(nside=nside)
    footprints = Footprint(mjd_start=survey_start, sun_ra_start=sun_ra_start, nside=nside)
    for f in footprints_hp.dtype.names:
        footprints.set_footprint(f, footprints_hp[f])

    if mask_basis_functions is None:
        mask_basis_functions = standard_masks(nside=nside)
    # Mask basis functions have zero weights.
    mask_basis_functions_weights = [0 for mask in mask_basis_functions]

    if reward_basis_functions is None:
        # Create list of (tuples of) basic reward basis functions and weights.
        m5_weight = 6.0
        footprint_weight = 1.5
        slewtime_weight = 3.0
        stayband_weight = 3.0
        repeat_weight = -5

        reward_functions = simple_rewards(
            footprints=footprints,
            bandname=bandname,
            nside=nside,
            m5_weight=m5_weight,
            footprint_weight=footprint_weight,
            slewtime_weight=slewtime_weight,
            stayband_weight=stayband_weight,
            repeat_weight=repeat_weight,
        )

        # unpack the basis functions and weights
        reward_basis_functions_weights = [val[1] for val in reward_functions]
        reward_basis_functions = [val[0] for val in reward_functions]

    # Set up survey name, use also for scheduler note.
    survey_name = f"simple greedy {bandname}"

    # Set up detailers for each requested observation.
    detailer_list = []
    # Avoid camera rotator limits.
    detailer_list.append(
        detailers.CameraRotDetailer(
            min_rot=np.min(camera_rot_limits),
            max_rot=np.max(camera_rot_limits),
            dither=camera_dither,
        )
    )
    # Convert rotTelPos to rotSkyPos_desired
    detailer_list.append(detailers.Rottep2RotspDesiredDetailer(telescope="rubin"))
    # Reorder visits in a blob so that closest to current altitude is first.
    detailer_list.append(detailers.CloseAltDetailer())
    # Sets a flush-by date to avoid running into prescheduled visits.
    detailer_list.append(detailers.FlushForSchedDetailer())

    # Set up the survey.
    ignore_obs = ["DD"]

    GreedySurvey_params = {
        "nside": nside,
        "seed": 42,
        "dither": dither,
    }

    greedy_survey = GreedySurvey(
        reward_basis_functions + mask_basis_functions,
        reward_basis_functions_weights + mask_basis_functions_weights,
        bandname=bandname,
        exptime=exptime,
        survey_name=survey_name,
        ignore_obs=ignore_obs,
        nexp=nexp,
        detailers=detailer_list,
        science_program=science_program,
        observation_reason=observation_reason,
        **GreedySurvey_params,
    )

    # Tucking this here so we can look at how many observations
    # recorded for this survey and what was the last one.
    greedy_survey.extra_features["ObsRecorded"] = features.NObsCount()
    greedy_survey.extra_features["LastObs"] = features.LastObservation()

    return greedy_survey


def simple_rewards_field_survey(
    scheduler_note: str,
    nside: int = DEFAULT_NSIDE,
    sun_alt_limit: float = -12.0,
) -> list[basis_functions.BaseBasisFunction]:
    """Get some simple rewards to observe a field survey for a long period.

    Parameters
    ----------
    scheduler_note : `str`
        The scheduler note for the field survey.
        Typically this will be the same as the field name.
    nside : `int`
        The nside value for the healpix grid.
    sun_alt_limit : `float`, optional
        Value for the sun's altitude at which to allow observations to start
        (or finish).

    Returns
    -------
    bfs : `list` of `~.scheduler.basis_functions.BaseBasisFunction`
    """
    bfs = [
        basis_functions.NotTwilightBasisFunction(sun_alt_limit=sun_alt_limit),
        # Avoid revisits within 30 minutes - but we'll have to replace "note"
        basis_functions.VisitGap(band_names=None, note=scheduler_note, gap_min=30.0),
        # reward fields which are rising, but don't mask out after zenith
        basis_functions.RewardRisingBasisFunction(nside=nside, slope=0.1, penalty_val=0),
        # Reward parts of the sky which are darker --
        # note that this is only for r band, so relying on skymap in r band.
        # if there isn't a strong reason to go with the darkest pointing,
        # it might be reasonable to just drop this basis function
        basis_functions.M5DiffBasisFunction(bandname="r", nside=nside),
    ]
    return bfs


def simple_field_survey(
    field_ra_deg: float,
    field_dec_deg: float,
    field_name: str,
    mask_basis_functions: list[basis_functions.BaseBasisFunction] | None = None,
    reward_basis_functions: list[basis_functions.BaseBasisFunction] | None = None,
    detailers: list[detailers.BaseDetailer] | None = None,
    sequence: str | list[str] = "ugrizy",
    nvisits: dict | None = None,
    exptimes: dict | None = None,
    nexps: dict | None = None,
    nside: int = DEFAULT_NSIDE,
    science_program: str | None = None,
) -> FieldSurvey:
    """Set up a simple field survey.

    Parameters
    ----------
    field_ra_deg : `float`
        The RA (in degrees) of the field.
    field_dec_deg : `float`
        The Dec (in degrees) of the field.
    field_name : `str`
        The name of the field. This is used for the survey_name and
        transferred to the 'target' information in the output observation.
        Also used in 'scheduler_note', which is important for the FieldSurvey
        to know whether to count particular observations for the Survey.
    mask_basis_functions : `list` [`BaseBasisFunction`] or None
        List of basis functions to use as masks (with implied weight 0).
        If None, `standard_masks` is used with default parameters.
    reward_basis_functions : `list` [`BaseBasisFunction`] or None
        List of basis functions to use as rewards.
        If None, a basic set of basis functions useful for long observations
        of a field within a night will be used (`get
    detailers : `list` of [`~.scheduler.detailer` objects]
        Detailers for the survey.
        Detailers can add information to output observations, including
        specifying rotator or dither positions.
    sequence : `str` or `list` [`str`]
        The bands (in order?) for the sequence of observations.
    nvisits : `dict` {`str`:`int`} | None
        Number of visits per band to program in the sequence.
        Default of None uses
        nvisits = {"u": 20, "g": 20, "r": 20, "i": 20, "z": 20, "y": 20}
    exptimes : `dict` {`str`: `float`} | None
        Exposure times per band to program in the sequence.
        Default of None uses
        exptimes = {"u": 38, "g": 30, "r": 30, "i": 30, "z": 30, "y": 30}
    nexps : `dict` {`str`: `int`} | None
        Number of exposures per band to program in the sequence.
        Default of None uses
        nexps = {"u": 1, "g": 2, "r": 2, "i": 2, "z": 2, "y": 2}
    nside : `int`, optional
        Nside for the survey. Default DEFAULT_NSIDE.
    science_program : `str` | None
        The science_program key for the FieldSurvey.
        This maps to the BLOCK and `science_program` in the consDB.

    Returns
    -------
    field_survey : `~.scheduler.surveys.FieldSurvey`
        The configured FieldSurvey.

    Notes
    -----
    The sequences for a given field survey can be set via kwargs,
    not necessarily easily accessible here. Only portions of the sequence
    which correspond to mounted bands will be requested by the FieldSurvey.

    field_survey.extra_features['ObsRecord'] tracks how many observations
    have been accepted by the Survey (and can be useful for diagnostics).
    """
    if mask_basis_functions is None:
        mask_basis_functions = standard_masks(nside=nside)
    if reward_basis_functions is None:
        reward_basis_functions = simple_rewards_field_survey(field_name, nside=nside)
    basis_functions = mask_basis_functions + reward_basis_functions

    if nvisits is None:
        nvisits = {"u": 20, "g": 20, "r": 20, "i": 20, "z": 20, "y": 20}
    if exptimes is None:
        exptimes = {"u": 38, "g": 30, "r": 30, "i": 30, "z": 30, "y": 30}
    if nexps is None:
        nexps = {"u": 1, "g": 2, "r": 2, "i": 2, "z": 2, "y": 2}

    field_survey = FieldSurvey(
        basis_functions,
        field_ra_deg,
        field_dec_deg,
        sequence=sequence,
        nvisits=nvisits,
        exptimes=exptimes,
        nexps=nexps,
        ignore_obs=None,
        survey_name=field_name,
        scheduler_note=field_name,
        target_name=field_name,
        readtime=2.4,
        band_change_time=120.0,
        nside=nside,
        flush_pad=30.0,
        detailers=detailers,
        science_program=science_program,
    )
    return field_survey



__all__ = ("FootprintBasisFunction",)

import warnings

import numpy as np

from rubin_scheduler.scheduler import features, utils
from rubin_scheduler.scheduler.basis_functions import BaseBasisFunction
from rubin_scheduler.scheduler.utils import ConstantFootprint
from rubin_scheduler.utils import DEFAULT_NSIDE


def send_unused_deprecation_warning(name):
    message = (
        f"The basis function {name} is not in use by the current "
        "baseline scheduler and may be deprecated shortly. "
        "Please contact the rubin_scheduler maintainers if "
        "this is in use elsewhere."
    )
    warnings.warn(message, FutureWarning)


class FootprintBasisFunction(BaseBasisFunction):
    """Basis function that tries to maintain a uniformly covered footprint

    Parameters
    ----------
    bandname : `str`, optional
        The band for this footprint. Default r.
    nside : `int`, optional
        The nside for the basis function and features.
        Default None uses `~utils.set_default_nside()`
    footprint : `~rubin_scheduler.scheduler.utils.Footprint` object
        The desired footprint. The default will set this to None,
        but in general this is really not desirable.
        In order to make default a kwarg, a current baseline footprint
        is set up with a Constant footprint (not rolling, not even season
        aware).
    out_of_bounds_val : `float`, optional
        The value to set the basis function for regions that are not in
        the footprint. Default -10, np.nan is another good value to use.
    seeing_fwhm_max : `float`
        Seeing limit to use (arcsec) when counting observations.
        Default None.
    """

    def __init__(
        self,
        bandname="r",
        nside=DEFAULT_NSIDE,
        footprint=None,
        out_of_bounds_val=-10.0,
        filtername=None,
        seeing_fwhm_max=None,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super().__init__(nside=nside, bandname=bandname)
        if footprint is None:
            # This is useful as a backup, but really footprint SHOULD
            # be specified when basis function is set up.
            # This just uses whole survey, but doesn't set up rolling.
            warnings.warn("No Footprint set, using a constant default.")
            target_maps, labels = utils.get_current_footprint(self.nside)
            fp = ConstantFootprint(nside=self.nside)
            for f in "ugrizy":
                fp.set_footprint(f, target_maps[f])
        self.footprint = footprint

        self.survey_features = {}
        # All the observations in all bands
        self.survey_features["N_obs_all"] = features.NObservations(
            nside=self.nside, bandname=None, seeing_limit=seeing_fwhm_max
        )
        self.survey_features["N_obs"] = features.NObservations(
            nside=self.nside, bandname=bandname, seeing_limit=seeing_fwhm_max
        )

        # should probably actually loop over all the target maps?
        self.out_of_bounds_area = np.where(footprint.get_footprint(self.bandname) <= 0)[0]
        self.out_of_bounds_val = out_of_bounds_val

    def _calc_value(self, conditions, indx=None):
        # Find out what the footprint object thinks we should have been
        # observed
        desired_footprint_normed = self.footprint(conditions.mjd)[self.bandname]

        # Compute how many observations we should have on the sky
        desired = desired_footprint_normed * np.sum(self.survey_features["N_obs_all"].feature)
        result = desired - self.survey_features["N_obs"].feature
        result[self.out_of_bounds_area] = self.out_of_bounds_val
        return result



__all__ = ("generate_ddf_grid",)

import argparse
import os
import sys

import astropy.units as u
import numpy as np
from astropy.time import Time

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.site_models.seeing_model import SeeingModel
from rubin_scheduler.utils import SURVEY_START_MJD, Site, ddf_locations, m5_flat_sed


def generate_ddf_grid(
    verbose=True,
    mjd0=59560.2,
    delta_t=15.0,
    survey_length=40.0,
    sun_limit=-12,
    nominal_expt=30.0,
):
    """Pre-compute conditions for DDF locations over survey

    Parameters
    ----------
    mjd0 : `float`
        The start MJD of the grid
    delta_t : `float`
        Spacing of time steps in minutes. Default 15
    survey_length : `float`
        Full span of DDF grid (years). Default 40.
    sun_limit : `float`
        Ignore times with sun above sun limit in degrees.
        Default -12.
    nominal_expt : `float`
        Nominal exposure time in seconds to use for depth visits.
        Default 30
    """

    # Technically this script should be over in rubin_sim, but here to be more
    # easily found. Bury import here so it's hopefully not a problem.
    import rubin_sim.skybrightness as sb
    from astroplan import Observer

    dds = ddf_locations()
    delta_t = delta_t / 60.0 / 24.0  # to days
    survey_length = survey_length * 365.25
    sun_limit = np.radians(sun_limit)  # degrees

    nominal_seeing = 0.7  # arcsec

    site = Site("LSST")
    observer = Observer(
        longitude=site.longitude * u.deg,
        latitude=site.latitude * u.deg,
        elevation=site.height * u.m,
        name="LSST",
    )

    seeing_model = SeeingModel()

    seeing_indx = 1  # 0=u, 1=g, 2=r, etc.

    mjds = np.arange(mjd0, mjd0 + survey_length, delta_t)

    names = ["mjd", "sun_alt", "sun_n18_rising_next", "moon_phase", "moon_alt"]
    for survey_name in dds.keys():
        names.append(survey_name + "_airmass")
        names.append(survey_name + "_sky_g")
        names.append(survey_name + "_m5_g")

    types = [float] * len(names)
    result = np.zeros(mjds.size, dtype=list(zip(names, types)))
    result["mjd"] = mjds

    # pretty sure these are radians
    ras = np.radians(np.array([dds[survey][0] for survey in dds]))
    decs = np.radians(np.array([dds[survey][1] for survey in dds]))

    sm = sb.SkyModel(mags=True)
    mags = []
    airmasses = []
    sun_alts = []
    moon_alts = []
    moon_phases = []

    maxi = mjds.size
    for i, mjd in enumerate(mjds):
        if verbose:
            progress = i / maxi * 100
            text = "\rprogress = %0.1f%%" % progress
            sys.stdout.write(text)
            sys.stdout.flush()

        try:
            sm.set_ra_dec_mjd(ras, decs, mjd, degrees=False)
        except ValueError:
            sm.sun_alt = 12.0
        if sm.sun_alt > sun_limit:
            mags.append(sm.return_mags()["g"] * 0)
            airmasses.append(sm.airmass * 0)
        else:
            mags.append(sm.return_mags()["g"])
            airmasses.append(sm.airmass)
        sun_alts.append(sm.sun_alt)
        result["sun_n18_rising_next"][i] = observer.twilight_morning_astronomical(
            Time(mjd, format="mjd"), which="next"
        ).mjd
        moon_alts.append(sm.moon_alt)
        moon_phases.append(sm.moon_phase)

    mags = np.array(mags)
    airmasses = np.array(airmasses)
    result["sun_alt"] = sun_alts
    result["moon_alt"] = moon_alts
    result["moon_phase"] = moon_phases

    for i, survey_name in enumerate(dds.keys()):
        result[survey_name + "_airmass"] = airmasses[:, i]
        result[survey_name + "_sky_g"] = mags[:, i]

        # now to compute the expected seeing if the zenith is nominal
        FWHMeff = seeing_model(nominal_seeing, airmasses[:, i])["fwhmEff"][seeing_indx, :]
        result[survey_name + "_m5_g"] = m5_flat_sed(
            "g", mags[:, i], FWHMeff, nominal_expt, airmasses[:, i], nexp=1
        )

    return result


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--full",
        dest="full",
        action="store_true",
        help="Compute 40 year grid at 15 min resolution. Default is 12 years at 5 min resolution.",
    )
    parser.set_defaults(full=False)

    args = parser.parse_args()

    full = args.full

    if full:
        result = generate_ddf_grid()
        np.savez(os.path.join(get_data_dir(), "scheduler", "ddf_grid.npz"), ddf_grid=result)
    else:
        result = generate_ddf_grid(verbose=True, mjd0=SURVEY_START_MJD - 365, delta_t=5.0, survey_length=12.0)
        np.savez(os.path.join(get_data_dir(), "scheduler", "ddf_grid_fine.npz"), ddf_grid=result)



__all__ = ("BandSwapScheduler", "SimpleBandSched", "ComCamBandSched", "BandSchedUzy", "DateSwapBandScheduler")

import numpy as np
from astropy.time import Time

from rubin_scheduler.scheduler.utils import IntRounded


# Dictionary keys based on astropy times make me nervous.
# Let's use the dayobs's string instead.
def time_to_key(time):
    return time.isot.split("T")[0]


def key_to_time(key):
    return Time(f"{key}T12:00:00", scale="tai")


class BandSwapScheduler:
    """A simple way to schedule what band to load"""

    def __init__(self):
        pass

    def add_observation(self, observation):
        pass

    def __call__(self, conditions):
        """
        Returns
        -------
        list of strings for the bands that should be loaded
        """
        pass


class SimpleBandSched(BandSwapScheduler):
    """Swap the mounted bands depending on the lunar phase.

    This assumes we swap between just two sets of bandpasses, one
    for lunar phases where moon illumination at sunset < `illum_limit` and
    another for phases > `illum_limit`.

    Parameters
    ----------
    illum_limit
        The illumination limit to be compared to the `moon_illum`
        reported by the `Almanac`.
    """

    def __init__(self, illum_limit=10.0):
        self.illum_limit_ir = IntRounded(illum_limit)

    def __call__(self, conditions):
        if IntRounded(conditions.moon_phase_sunset) > self.illum_limit_ir:
            result = ["g", "r", "i", "z", "y"]
        else:
            result = ["u", "g", "r", "i", "z"]
        return result


class ComCamBandSched(BandSwapScheduler):
    """ComCam can only hold 3 bands at a time.

    Pretend we will cycle from ugr, gri, riz, izy
    depending on lunar phase.

    Parameters
    ----------
    loaded_band_groups : `tuple` (`tuple`)
        Groups of 3 bands, to be loaded at the same time.
        Multiple groups can be specified, to be swapped between at the
        boundaries of `illum_bins` in lunar phase.
    illum_bins : `np.ndarray`, (N,)
        Lunar illumination boundaries to define when to swap between
        different groups of bands within the `loaded_band_groups`.
        Lunar illumination ranges from 0 to 100.

    Notes
    -----
    If illum_bins = np.array([0, 50, 100]), then there should be
    two groups of bands to use -- one for use between 0 and 50 percent
    illumination, and another for use between 50 and 100 percent illumination.
    """

    def __init__(
        self,
        loaded_band_groups=(("u", "g", "r"), ("g", "r", "i"), ("r", "i", "z"), ("i", "z", "y")),
        illum_bins=np.arange(0, 100 + 1, 25),
    ):
        self.loaded_band_groups = loaded_band_groups
        self.illum_bins = illum_bins
        if isinstance(self.illum_bins, list):
            self.illum_bins = np.array(illum_bins)
        if len(illum_bins) - 1 > len(loaded_band_groups):
            raise ValueError("There are illumination bins with an " "undefined loaded_band_group")

    def __call__(self, conditions):
        moon_at_sunset = conditions.moon_phase_sunset
        try:
            if len(moon_at_sunset) > 0:
                moon_at_sunset = moon_at_sunset[0]
        except TypeError:
            pass
        indx = np.searchsorted(self.illum_bins, moon_at_sunset, side="left")
        indx = np.max([0, indx - 1])
        result = list(self.loaded_band_groups[indx])
        return result


class BandSchedUzy(BandSwapScheduler):
    """
    remove u in bright time. Alternate between removing z and y in
    dark time.

    Note, this might not work properly if we need to restart a bunch.
    So a more robust way of scheduling band loading might be in order.
    """

    def __init__(self, illum_limit=10.0):
        self.illum_limit_ir = IntRounded(illum_limit)
        self.last_swap = 0

        self.bright_time = ["g", "r", "i", "z", "y"]
        self.dark_times = [["u", "g", "r", "i", "y"], ["u", "g", "r", "i", "z"]]

    def __call__(self, conditions):
        if IntRounded(conditions.moon_phase_sunset) > self.illum_limit_ir:
            result = self.bright_time
        else:
            indx = self.last_swap % 2
            result = self.dark_times[indx]
            if result != conditions.mounted_bands:
                self.last_swap += 1
        return result


class DateSwapBandScheduler(BandSwapScheduler):
    """Swap specific bands on specific days, up until end_date, then
    fall back to the backup_filter_scheduler.

    This provides a way for simulations to use the specific bands that
    were in use on specific nights, as well as to incorporate knowledge
    of the upcoming scheduler, while still falling back to a reasonable
    filter swap schedule beyond the time the dates are precisely known.

    Parameters
    ----------
    swap_schedule
        Dictionary of times of the filter swaps, together with the filter
        complement loaded into the carousel.
        e.g. {"2025-08-20" : ['r', 'i', 'z', 'y']}
    end_date
        The Time after which the dictionary is no longer valid and the
        `backup_band_scheduler` should be used.
    backup_band_scheduler
        The more general band swap scheduler to use after the specific
        dates of the dictionary are exhausted. This should match the
        expected band scheduler for the Scheduler.
    """

    def __init__(
        self,
        swap_schedule: dict[str, list[str]] | None = None,
        end_date: Time | None = None,
        backup_band_scheduler: BandSwapScheduler = SimpleBandSched(illum_limit=40),
    ):
        previous_swap_schedule = {
            "2025-06-20": ["u", "g", "r", "i", "z"],
            "2025-07-01": ["g", "r", "i", "z"],
            "2025-07-04": ["g", "r", "i", "z", "y"],
            "2025-07-10": ["z"],
            "2025-07-11": ["g", "r", "i", "z", "y"],
            "2025-07-15": ["u", "g", "r", "i", "z"],
            "2025-07-28": [
                "u",
                "r",
                "i",
                "z",
            ],
            "2025-08-07": ["r", "i", "z", "y"],
            "2025-08-12": ["g", "r", "i", "z"],
        }
        previous_swap_times = np.sort(np.array([key_to_time(k) for k in previous_swap_schedule.keys()]))

        if swap_schedule is None:
            # Current estimate for the SV survey.
            # Subject to change, although past dates should match reality.
            swap_schedule = {
                "2025-08-12": ["g", "r", "i", "z"],
            }
        new_swap_times = np.sort(np.array([key_to_time(k) for k in swap_schedule.keys()]))

        # Join previous_swap_schedule and new schedule -
        # Use new swap_schedule if time overlaps with previous.
        keep_times = np.where(previous_swap_times < new_swap_times.min())[0]
        keep_keys = [time_to_key(t) for t in previous_swap_times[keep_times]]
        previous_swap_schedule = dict([(k, previous_swap_schedule[k]) for k in keep_keys])

        self.swap_schedule = previous_swap_schedule
        self.swap_schedule.update(swap_schedule)
        self.swap_schedule_times = np.sort(np.array([key_to_time(k) for k in self.swap_schedule.keys()]))

        if end_date is None:
            self.end_date = Time("2025-09-25T12:00:00")
        else:
            self.end_date = end_date

        if backup_band_scheduler is None:
            self.backup_band_scheduler = SimpleBandSched(illum_limit=40)
        else:
            self.backup_band_scheduler = backup_band_scheduler

    def __call__(self, conditions):
        current_time = Time(conditions.mjd, format="mjd", scale="tai")

        # Are we within the bounds of the scheduled swaps?
        if current_time < self.end_date:
            idx = np.where(current_time >= self.swap_schedule_times)[0][-1]
            return self.swap_schedule[time_to_key(self.swap_schedule_times[idx])]

        else:
            return self.backup_band_scheduler(conditions)



__all__ = (
    "BaseDetailer",
    "ZeroRotDetailer",
    "Comcam90rotDetailer",
    "Rottep2RotspDesiredDetailer",
    "CloseAltDetailer",
    "TakeAsPairsDetailer",
    "TwilightTripleDetailer",
    "FlushForSchedDetailer",
    "FilterNexp",
    "BandNexp",
    "FixedSkyAngleDetailer",
    "ParallacticRotationDetailer",
    "FlushByDetailer",
    "RandomFilterDetailer",
    "RandomBandDetailer",
    "BandSortDetailer",
    "TruncatePreTwiDetailer",
    "TrackingInfoDetailer",
    "AltAz2RaDecDetailer",
    "StartFieldSequenceDetailer",
    "BandToFilterDetailer",
    "TagRadialDetailer",
    "CopyValueDetailer",
    "LabelRegionDetailer",
    "LabelDDFDetailer",
    "LabelRegionsAndDDFs",
    "RollBandMatchDetailer",
)

import copy
import warnings

import healpy as hp
import numpy as np

import rubin_scheduler.scheduler.features as features
from rubin_scheduler.scheduler.utils import CurrentAreaMap, HpInLsstFov, IntRounded, ObservationArray
from rubin_scheduler.utils import (
    DEFAULT_NSIDE,
    _angular_separation,
    _approx_alt_az2_ra_dec,
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    ddf_locations,
    pseudo_parallactic_angle,
    rotation_converter,
)


class BaseDetailer:
    """
    A Detailer is an object that takes a list of proposed observations and
    adds "details" to them. The primary purpose is that the Markov Decision
    Process does an excelent job selecting RA,Dec,band combinations, but we
    may want to add additional logic such as what to set the camera rotation
    angle to, or what to use for an exposure time. We could also modify the
    order of the proposed observations. For Deep Drilling Fields, a detailer
    could be useful for computing dither positions and modifying the exact
    RA,Dec positions.
    """

    def __init__(self, nside=DEFAULT_NSIDE):
        """"""
        # Dict to hold all the features we want to track
        self.survey_features = {}
        self.nside = nside

    def add_observations_array(self, observations_array, observations_hpid):
        """Like add_observation, but for loading a whole array of
        observations at a time"""
        if hasattr(self, "survey_features"):
            for feature in self.survey_features:
                self.survey_features[feature].add_observations_array(observations_array, observations_hpid)

    def add_observation(self, observation, indx=None):
        """
        Parameters
        ----------
        observation : `np.array`
            An array with information about the input observation
        indx : `np.array`
            The indices of the healpix map that the observation overlaps
            with
        """
        if hasattr(self, "survey_features"):
            for feature in self.survey_features:
                self.survey_features[feature].add_observation(observation, indx=indx)

    def __call__(self, observation_list, conditions):
        """
        Parameters
        ----------
        observation_list : `list` of observations
            The observations to detail.
        conditions : `rubin_scheduler.scheduler.conditions` object

        Returns
        -------
        ObservationArray
        """

        return ObservationArray()


class TrackingInfoDetailer(BaseDetailer):
    """Add metadata related to target_name, observation_reason and
    target_name for an observation.

    Does NOT clobber information that has been previously set.
    Appended to all surveys detailer lists, if not already present,
    by BaseSurvey using values from survey configuration.

    See https://rtn-096.lsst.io for further information on these observation
    metadata fields.

    Parameters
    ----------
    target_name : `str` or `None`
        The value to enter into the target_name field of the Observation.
        With current metadata expectations, this will indicate a specific
        target name or a general region label. See also LabelRegionDetailer.
    science_program : `str` or `None`
        The value to enter into the science_program field of the Observation.
        With the current configuration of the SchedulerCSC, this corresponds
        to the JSON BLOCK executed to acquire the visit.
    observation_reason : `str` or `None`
        The value to enter into the observation_reason of the Observation.
        With current metadata expectations, this will indicate something
        about the observing mode.
        Spaces in observation_reason will be replaced by underscores,
        required by camera handling of metadata.
    """

    def __init__(self, target_name="", science_program="", observation_reason=""):
        self.survey_features = {}
        self.target_name = target_name
        self.science_program = science_program
        self.observation_reason = observation_reason
        if self.observation_reason is not None:
            self.observation_reason = self.observation_reason.replace(" ", "_")
        self.keys = ["science_program", "target_name", "observation_reason"]

    def __call__(self, observation_array, conditions):
        for key in self.keys:
            indx = np.where(
                (observation_array[key] == "")
                | (observation_array[key] is None)
                | (observation_array[key] == "None")
            )[0]
            observation_array[key][indx] = getattr(self, key)

        return observation_array


class RollBandMatchDetailer(BaseDetailer):
    """Roll the order of visits to eliminate a filter change."""

    def __call__(self, observation_array, conditions):

        bm_index = np.where(observation_array["band"] == conditions.current_band)[0]
        if np.size(bm_index) > 0:
            indx = np.arange(observation_array.size)
            indx = np.roll(indx, -np.min(bm_index))
            observation_array = observation_array[indx]

        return observation_array


class FlushByDetailer(BaseDetailer):
    """Set the MJD an observation should be flushed from the scheduler
    queue if not yet completed.

    Parameters
    ----------
    flush_time : `float`
        The time to flush after the current MJD. Default 60 minutes
    """

    def __init__(self, flush_time=60):
        self.survey_features = {}
        self.flush_time = flush_time / 60 / 24.0

    def __call__(self, observation_array, conditions):
        observation_array["flush_by_mjd"] = conditions.mjd + self.flush_time
        return observation_array


class BandToFilterDetailer(BaseDetailer):
    """If we want to fill in the physical filter to request rather
    than just the band.

    Parameters
    ----------
    band_to_filter_dict : `dict`
        A dict that maps band name (usually ugrizy) to
        specific filter names. Default value of None
        will set the filter name to the same as the band name.
    """

    def __init__(self, band_to_filter_dict=None):
        self.survey_features = {}
        if band_to_filter_dict is None:
            self.band_to_filter_dict = {}
            for bandname in "ugrizy":
                self.band_to_filter_dict[bandname] = bandname
        else:
            self.band_to_filter_dict = band_to_filter_dict

    def __call__(self, observation_array, conditions):
        u_bands = np.unique(observation_array["band"])
        for band in u_bands:
            indx = np.where(observation_array["band"] == band)
            # Fetch the dictionary value or just continue to use band
            filtername = self.band_to_filter_dict.get(band, band)
            observation_array["filter"][indx] = filtername

        return observation_array


class TruncatePreTwiDetailer(BaseDetailer):
    """Truncate an array of observations so they fit before
    morning twilight starts.

    Parameters
    ----------
    pad : `float`
        The pad to give before start of morning twilight.
        Default 5 (minutes).
    filter_change_time : `float`
        Estimate of filter change time
    visit_overhead : `float`
        Estimate of visit overhead (read, slew, settle).
        Default 4.0 (seconds).
    twilight_cut_to : `str`
        Cut to sun altitude of -18 degrees or -12 degrees.
        Valid values of "n18" or "n12".
    """

    def __init__(self, pad=5.0, filter_change_time=120.0, visit_overhead=4.0, twilight_cut_to="n18"):
        self.pad = pad / 60.0 / 24.0  # To days
        self.filter_change_time = filter_change_time / 3600.0 / 24.0  # To days

        self.visit_overhead = visit_overhead / 3600.0 / 24.0  # To days
        self.twilight_cut_to = twilight_cut_to

    def __call__(self, observation_array, conditions):

        b1 = observation_array["band"][0:-1]
        b2 = observation_array["band"][1:]

        f_changes = b1 != b2
        # Do we start with a filter change?
        prepend = observation_array["band"][0] != conditions.current_band
        f_changes = np.array([prepend] + np.array(f_changes).tolist())
        f_time = f_changes * self.filter_change_time

        tot_time = observation_array["exptime"] / 3600.0 / 24.0 + f_time + self.visit_overhead
        cumulative_time = np.cumsum(tot_time)

        cut_mjd = getattr(conditions, "sun_%s_rising" % self.twilight_cut_to)

        time_avail = cut_mjd - conditions.mjd - self.pad

        indx = np.where(IntRounded(cumulative_time) <= IntRounded(time_avail))[0]
        if np.size(indx) > 0:
            trunc_indx = np.max(indx)
            return observation_array[0 : trunc_indx + 1]
        else:
            return ObservationArray(n=0)


class AltAz2RaDecDetailer(BaseDetailer):
    """Set RA,dec for an observation that only has alt,az"""

    def __call__(self, observation_array, conditions):
        ra, dec = _approx_alt_az2_ra_dec(
            observation_array["alt"],
            observation_array["az"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        observation_array["RA"] = ra
        observation_array["dec"] = dec

        return observation_array


class RandomFilterDetailer(BaseDetailer):
    """Deprecated in favor of RandomBandDetailer"""

    def __init__(self, filters="riz", nights_to_prep=10000, seed=42, fallback_order="rizgyu"):
        warnings.warn("Deprecated in favor of RandomBandDetailer", FutureWarning)
        super().__init__(
            bands=filters, nights_to_prep=nights_to_prep, seed=seed, fallback_order=fallback_order
        )


class RandomBandDetailer(BaseDetailer):
    """Pick a random band for the observations

    Parameters
    ----------
    bands : `str`
        The bands to randomize. Default 'riz'
    nights_to_prep : `int`
        The number of nights to generate random bands for.
        Default 10000.
    seed : number
        Seed for RNG. Defaut 42
    fallback_order : `str`
        If the desired band is not mounted, goes through
        `fallback_order` and uses the first band that is
        available
    """

    def __init__(self, bands="riz", nights_to_prep=10000, seed=42, fallback_order="rizgyu"):
        self.survey_features = []
        self.fallback_order = fallback_order

        rng = np.random.default_rng(seed)
        self.night2band_int = rng.integers(low=0, high=len(bands), size=nights_to_prep)

        self.band_dict = {}
        for i, bandname in enumerate(bands):
            self.band_dict[i] = bandname

    def __call__(self, observation_array, conditions):
        band_to_use = self.band_dict[self.night2band_int[conditions.night]]
        # Band not available
        if band_to_use not in conditions.mounted_bands:
            is_mounted = [bandname in conditions.mounted_bands for bandname in self.fallback_order]
            indx = np.min(np.where(is_mounted))
            band_to_use = self.fallback_order[indx]

        observation_array["band"] = band_to_use
        return observation_array


class BandSortDetailer(BaseDetailer):
    """Detailer that sorts an array of observations by specified
    band order, in order to minimize filter changes.

    Useful for scripted surveys with many filter changes like DDFs.

    Parameters
    ----------
    desired_band_order : `str`
        The desired band order. Default of None retains order
        and only moves observations up if they match the current filter.
    loaded_first : `bool`
        If True, then the currently-in-use filter is always moved to
        the start of the desired band order, to remove the first filter change.
    """

    def __init__(
        self,
        desired_band_order: str = None,
        loaded_first: bool = True,
    ):
        self.desired_band_order = desired_band_order
        self.loaded_first = loaded_first

    def __call__(
        self, observation_array: ObservationArray, conditions: features.Conditions
    ) -> ObservationArray:

        order_to_set = copy.copy(self.desired_band_order)

        u_vals, indx = np.unique(observation_array["band"], return_index=True)
        # If we have only 1 filter, nothing to sort, so return
        if np.size(u_vals) == 1:
            return observation_array

        if order_to_set is None:
            # unique values, in order they appear, as a string
            order_to_set = "".join(u_vals[np.argsort(indx)].tolist())

        if self.loaded_first:
            order_to_set = order_to_set.replace(conditions.current_band, "")
            order_to_set = conditions.current_band + order_to_set
        indicies = []
        for bandname in order_to_set:
            indicies.append(np.where(observation_array["band"] == bandname)[0])

        indices = np.concatenate(indicies)

        return observation_array[indices]


class ParallacticRotationDetailer(BaseDetailer):
    """Set the rotator to near the parallactic angle"""

    def __init__(self, telescope="rubin"):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, observation_array, conditions, limits=[-270, 270]):
        limits = np.radians(limits)

        alt, az = _approx_ra_dec2_alt_az(
            observation_array["RA"],
            observation_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
        observation_array["rotSkyPos_desired"] = obs_pa

        resulting_rot_tel_pos = self.rc._rotskypos2rottelpos(observation_array["rotSkyPos_desired"], obs_pa)

        indx = np.where(resulting_rot_tel_pos > np.max(limits))[0]
        resulting_rot_tel_pos[indx] -= 2 * np.pi

        indx = np.where(resulting_rot_tel_pos < np.min(limits))
        resulting_rot_tel_pos[indx] += 2 * np.pi

        # If those corrections still leave us bad, just pull it back 180.
        indx = np.where(resulting_rot_tel_pos > np.max(limits))
        resulting_rot_tel_pos[indx] -= np.pi

        # The rotTelPos overides everything else.
        observation_array["rotTelPos"] = resulting_rot_tel_pos
        # if the rotSkyPos_desired isn't possible, fall back to this.
        observation_array["rotTelPos_backup"] = 0

        return observation_array


class Rottep2RotspDesiredDetailer(BaseDetailer):
    """Convert all the rotTelPos values to rotSkyPos_desired"""

    def __init__(self, telescope="rubin"):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, obs_array, conditions):
        alt, az = _approx_ra_dec2_alt_az(
            obs_array["RA"],
            obs_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)

        rot_sky_pos_desired = self.rc._rotskypos2rottelpos(obs_array["rotTelPos"], obs_pa)

        obs_array["rotTelPos_backup"] = obs_array["rotTelPos"] + 0
        obs_array["rotTelPos"] = np.nan
        obs_array["rotSkyPos"] = np.nan
        obs_array["rotSkyPos_desired"] = rot_sky_pos_desired

        return obs_array


class ZeroRotDetailer(BaseDetailer):
    """
    Detailer to set the camera rotation to be apporximately zero in
    rotTelPos.

    Parameters
    ----------
    telescope : `str`
        Which telescope convention to use for setting the conversion
        between rotTelPos and rotSkyPos. Default "rubin".
    """

    def __init__(self, telescope="rubin", nside=DEFAULT_NSIDE):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, observation_array, conditions):
        obs_pa, alt, az = pseudo_parallactic_angle(
            observation_array["RA"],
            observation_array["dec"],
            conditions.mjd,
            np.degrees(conditions.site.longitude_rad),
            np.degrees(conditions.site.latitude_rad),
        )

        observation_array["rotSkyPos"] = np.radians(self.rc.rottelpos2rotskypos(0.0, obs_pa))

        return observation_array


class Comcam90rotDetailer(BaseDetailer):
    """
    Detailer to set the camera rotation so rotSkyPos is 0, 90, 180, or
    270 degrees. Whatever is closest to rotTelPos of zero.
    """

    def __init__(self, telescope="rubin", nside=DEFAULT_NSIDE):
        self.rc = rotation_converter(telescope=telescope)
        self.survey_features = {}

    def __call__(self, obs_array, conditions):
        favored_rot_sky_pos = np.radians([0.0, 90.0, 180.0, 270.0, 360.0]).reshape(5, 1)

        parallactic_angle, alt, az = np.radians(
            pseudo_parallactic_angle(
                obs_array["RA"],
                obs_array["dec"],
                conditions.mjd,
                np.degrees(conditions.site.longitude_rad),
                np.degrees(conditions.site.latitude_rad),
            )
        )

        # need to find the

        ang_diff = np.abs(self.rc._rotskypos2rottelpos(favored_rot_sky_pos, np.radians(parallactic_angle)))
        min_indxs = np.argmin(ang_diff, axis=0)
        # can swap 360 and zero if needed?
        obs_array["rotSkyPos"] = favored_rot_sky_pos[min_indxs].ravel()

        return obs_array


class StartFieldSequenceDetailer(BaseDetailer):
    """Prepend a sequence of observations to the start of an array

    Parameters
    ----------
    sequence_obs : `ObservationArray`
        ObservationArray object. The observations should
        have "scheduler_note" and/or "science_program" set.
    ang_distance_match : `float`
        How close should an observation be on the sky to be considered
        matching (degrees).
    time_match_hours : `float`
        How close in time to demand an observation be matching (hours).
    science_program : `str` or None
        The science_program to match against. Default None.
    scheduler_note : `str` or None
        The scheduler_note to match observations against.
        Default "starting_sequence".
    ra : `float`
        RA to match against. Default 0 (degrees). Ignored
        if ang_distance_match is None.
    dec : `float`
        Dec to match observations against. Default 0 (degrees).
        Ignored if ang_distance_match is None.

    """

    def __init__(
        self,
        sequence_obs,
        ang_distance_match=3.5,
        time_match_hours=5,
        science_program=None,
        scheduler_note="starting_sequence",
        ra=0,
        dec=0,
    ):
        super().__init__()
        self.survey_features["last_matching"] = features.LastObservedMatching(
            ang_distance_match=3.5,
            science_program=science_program,
            scheduler_note=scheduler_note,
            ra=ra,
            dec=dec,
        )
        self.ang_distance_match = np.radians(ang_distance_match)
        self.time_match = time_match_hours / 24.0
        self.science_program = science_program
        self.scheduler_note = scheduler_note

        # Make backwards compatible if someone sent in a list
        if isinstance(sequence_obs, list):
            warnings.warn("sequence_obs should be ObsArray, not list of ObsArray. Concatenating")
            sequence_obs = np.concatenate(sequence_obs)

        self.sequence_obs = sequence_obs

        self.sequence_obs["science_program"] = self.science_program
        self.sequence_obs["scheduler_note"] = self.scheduler_note

        # Check that things are sensibly set

        u_scip = np.unique(sequence_obs["science_program"])
        u_sn = np.unique(sequence_obs["scheduler_note"])

        if (np.size(u_scip) > 1) | (np.size(u_sn) > 1):
            msg = (
                "The science_program and/or scheduler_note " "values in sequence_obs_list should be the same."
            )
            raise ValueError(msg)

        if science_program is not None:
            if self.science_program != u_scip:
                ValueError("science_program kwarg not equal to science_programs from sequence_obs_list")
        if scheduler_note is not None:
            if self.scheduler_note != u_sn:
                ValueError("scheduler_note kwarg not equal to scheduler_notes from sequence_obs_list")

    def __call__(self, observation_array, conditions):
        # Do we need to add the opening sequence?
        if (conditions.mjd - self.survey_features["last_matching"].feature["mjd"]) >= self.time_match:
            observation_array = np.concatenate([self.sequence_obs, observation_array])

        return observation_array


class FixedSkyAngleDetailer(BaseDetailer):
    """Detailer to force a specific sky angle.

    Parameters
    ----------
    sky_angle : `float`, optional
        Desired sky angle (default = 0, in degrees).
    """

    def __init__(self, sky_angle=0.0, nside=DEFAULT_NSIDE):
        super().__init__(nside=nside)

        self.sky_angle = np.radians(sky_angle)

    def __call__(self, observation_array, conditions):
        observation_array["rotSkyPos"] = self.sky_angle

        return observation_array


class CloseAltDetailer(BaseDetailer):
    """Re-order a list of observations so that the closest in altitude to
    the current pointing is first.

    Parameters
    ----------
    alt_band : `float` (10)
        The altitude band to try and stay in (degrees)
    """

    def __init__(self, alt_band=10.0):
        super(CloseAltDetailer, self).__init__()
        self.alt_band = IntRounded(np.radians(alt_band))

    def __call__(self, obs_array, conditions):
        alt, az = _approx_ra_dec2_alt_az(
            obs_array["RA"],
            obs_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        alt_diff = np.abs(alt - conditions.tel_alt)
        in_band = np.where(IntRounded(alt_diff) <= self.alt_band)[0]
        if in_band.size == 0:
            in_band = np.arange(alt.size)

        # Find the closest in angular distance of the points that are in band
        ang_dist = _angular_separation(az[in_band], alt[in_band], conditions.tel_az, conditions.tel_alt)
        if np.size(ang_dist) == 1:
            good = 0
        else:
            good = np.min(np.where(ang_dist == ang_dist.min())[0])
        indx = in_band[good]
        result = np.concatenate([obs_array[indx:], obs_array[:indx]])
        return result


class FlushForSchedDetailer(BaseDetailer):
    """Update the flush-by MJD to be before any scheduled observations

    Parameters
    ----------
    tol : `float`
         How much before to flush (minutes)
    """

    def __init__(self, tol=2.5):
        super(FlushForSchedDetailer, self).__init__()
        self.tol = tol / 24.0 / 60.0  # To days

    def __call__(self, observation_array, conditions):
        if np.size(conditions.scheduled_observations) > 0:
            new_flush = np.min(conditions.scheduled_observations) - self.tol
            indx = np.where(observation_array["flush_by_mjd"] > new_flush)[0]
            observation_array[indx]["flush_by_mjd"] = new_flush
        return observation_array


class BandNexp(BaseDetailer):
    """Demand one band always be taken as a certain number of exposures"""

    def __init__(self, bandname="u", nexp=1, exptime=None):
        super(BandNexp, self).__init__()
        self.bandname = bandname
        self.nexp = nexp
        self.exptime = exptime

    def __call__(self, observation_array, conditions):
        indx = np.where(observation_array["band"] == self.bandname)[0]
        observation_array["nexp"][indx] = self.nexp
        if self.exptime is not None:
            observation_array["exptime"][indx] = self.exptime

        return observation_array


class FilterNexp(BandNexp):
    """Deprecated in favor of BandNexp"""

    def __init__(self, filtername="u", nexp=1, exptime=None):
        warnings.warn("Deprecated in favor of BandNexp", FutureWarning)
        super().__init__(bandname=filtername, nexp=nexp, exptime=exptime)


class TakeAsPairsDetailer(BaseDetailer):
    def __init__(self, bandname="r", exptime=None, nexp_dict=None, filtername=None):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super(TakeAsPairsDetailer, self).__init__()
        self.bandname = bandname
        self.exptime = exptime
        self.nexp_dict = nexp_dict

    def __call__(self, observation_array, conditions):
        paired = copy.deepcopy(observation_array)
        if self.exptime is not None:
            paired["exptime"] = self.exptime
        paired["band"] = self.bandname
        if self.nexp_dict is not None:
            paired["nexp"] = self.nexp_dict[self.bandname]
        if conditions.current_band == self.bandname:
            tags = ["a", "b"]
        else:
            tags = ["b", "a"]

        paired["scheduler_note"] = np.char.add(paired["scheduler_note"], ", %s" % tags[0])
        observation_array["scheduler_note"] = np.char.add(
            observation_array["scheduler_note"], ", %s" % tags[1]
        )

        # Try to avoid extra filter changes and have "a" tag
        # come first.
        if conditions.current_band == self.bandname:
            result = np.concatenate([paired, observation_array])
        else:
            result = np.concatenate([observation_array, paired])

        return result


class TwilightTripleDetailer(BaseDetailer):
    def __init__(self, slew_estimate=5.0, n_repeat=3, update_note=True):
        super(TwilightTripleDetailer, self).__init__()
        self.slew_estimate = slew_estimate
        self.n_repeat = n_repeat
        self.update_note = update_note

    def __call__(self, obs_array, conditions):
        # Estimate how much time is left in the twilgiht block
        potential_times = np.array(
            [
                conditions.sun_n18_setting - conditions.mjd,
                conditions.sun_n12_rising - conditions.mjd,
            ]
        )

        potential_times = np.min(potential_times[np.where(potential_times > 0)]) * 24.0 * 3600.0

        # How long will observations take?
        cumulative_slew = np.arange(obs_array.size) * self.slew_estimate
        cumulative_expt = np.cumsum(obs_array["exptime"])
        cumulative_time = cumulative_slew + cumulative_expt
        # If we are way over, truncate the list before doing the triple
        if np.max(cumulative_time) > potential_times:
            max_indx = np.where(cumulative_time / self.n_repeat <= potential_times)[0]
            if np.size(max_indx) == 0:
                # Very bad magic number fudge
                max_indx = 3
            else:
                max_indx = np.max(max_indx)
                if max_indx == 0:
                    max_indx += 1
            obs_array = obs_array[0:max_indx]

        # Repeat the observations n times
        out_obs = []
        for i in range(self.n_repeat):
            sub_arr = copy.deepcopy(obs_array)
            if self.update_note:
                sub_arr["scheduler_note"] = np.char.add(sub_arr["scheduler_note"], ", %i" % i)
            out_obs.append(sub_arr)
        return np.concatenate(out_obs)


class TagRadialDetailer(BaseDetailer):
    """Update target_name and scheduler note based on visit location.

    Parameters
    ----------
    radius : `float`
        How close to a location on the sky a visit must be to
        get tagged. Defualr 5.6 (degrees).
    ra : `float`
        RA of the location. Default 0 (degrees).
    dec : `float`
        Dec of the location. Default -30 (degrees).
    note_append : `str`
        The string to append to the visit scheduler_note.
        Default ", deep area".
    target_prefix : `str`
        String to add to the start of the target_name field.
        Default "deep "
    target_name : `str`
        Target name to use. Overrides name generater with target_prefix.
        Default None.
    """

    def __init__(
        self, radius=5.6, ra=0, dec=-30.0, note_append=", deep area", target_prefix="deep ", target_name=None
    ):
        self.survey_features = {}
        self.radius = np.radians(radius)
        self.ra = np.radians(ra)
        self.dec = np.radians(dec)
        self.note_append = note_append
        self.target_name = target_prefix + f"{np.degrees(self.ra):.3f} {np.degrees(self.dec):.3f}"
        if target_name is not None:
            self.target_name = target_name

    def __call__(self, obsarray, conditions):
        distances = _angular_separation(obsarray["RA"], obsarray["dec"], self.ra, self.dec)
        if np.size(distances) == 1:
            distances = np.array([distances])
        in_region = np.where(distances <= self.radius)[0]
        obsarray["target_name"][in_region] = self.target_name
        obsarray["scheduler_note"][in_region] = np.char.add(
            obsarray["scheduler_note"][in_region], self.note_append
        )
        return obsarray


class LabelRegionDetailer(BaseDetailer):
    """Label which region(s) of the footprint we are in.

    Parameters
    ----------
    label_array : `np.array`
        A HEALpix array of strings that are labels for each HEALpix.
        Default of None uses CurrentAreaMap to load labels.
    field_for_label : `str`
        Which ObservationArray field should be modified.
        Default "target_name".
    camera : `str`
        Which camera model to use to convert a pointing to
        HEALpix IDs. Default "LSST".
    append : `bool`
        Should the labels be appended to `field_for_label`
        (True, default), or clobber (False).
    separator : `str`
        If append is True, what string to use when appending.
        Default ", ".
    """

    def __init__(
        self, label_array=None, field_for_label="target_name", camera="LSST", append=True, separator=", "
    ):
        if label_array is None:
            sky = CurrentAreaMap()
            _footprints, self.label_array = sky.return_maps()
        else:
            self.label_array = label_array
        self.field_for_label = field_for_label
        nside = hp.npix2nside(np.size(self.label_array))
        self.append = append
        self.separator = separator

        self.remove_vals = set(["", "", "None"])

        if camera == "LSST":
            self.pointing2hpindx = HpInLsstFov(nside=nside)
        else:
            raise ValueError("Unknown camera")

    def __call__(self, obs_array, conditions):
        for i in np.arange(obs_array.size):
            indx = self.pointing2hpindx(obs_array["RA"][i], obs_array["dec"][i])
            labels = np.unique(self.label_array[indx])
            # ignore things that are out of bounds
            labels = labels[np.where(labels != "")]
            # If there are no applicable labels, identify as OutOfFootprint
            if np.size(labels) == 0:
                result = set(["OOF"])
            else:
                result = set(labels)

            if self.append:
                prev_values = set(obs_array[self.field_for_label][i].split(self.separator))
                new_values = prev_values.union(result).difference(self.remove_vals)
            else:
                new_values = result.difference(self.remove_vals)
            obs_array[self.field_for_label][i] = self.separator.join(new_values)
        return obs_array


class LabelDDFDetailer(BaseDetailer):
    """Label if an observation is close enough to a
    DDF location to be tagged.

    Parameters
    ----------
    ddf_locations : `dict`
        Dictionary with keys of DDF names and values of RA,dec
        pairs (in degrees). Default of None uses ddf_locations
        utility to get default locations and names.
    field_for_label : `str`
        Which ObservationArray field should be modified.
        Default "target_name".
    append : `bool`
        Should the labels be appended to `field_for_label`
        (True, default), or clobber (False).
    match_radius : `float`
        The radius away an observation can be from a DDF center
        and still be tagged. Default 2.0 (degrees)
    separator : `str`
        If append is True, what string to use when appending.
        Default ", ".
    """

    def __init__(
        self, ddf_location=None, field_for_label="target_name", append=True, match_radius=2.0, separator=", "
    ):
        self.field_for_label = field_for_label
        self.append = append
        self.match_radius = np.radians(match_radius)
        self.separator = separator
        if ddf_location is None:
            self.ddf_locations = ddf_locations()
        else:
            self.ddf_locations = ddf_locations

        for key in self.ddf_locations:
            self.ddf_locations[key] = np.radians(self.ddf_locations[key])

        self.remove_vals = set(["", "", "None"])

    def __call__(self, obs_array, conditions):

        for name in self.ddf_locations:
            distances = _angular_separation(
                self.ddf_locations[name][0], self.ddf_locations[name][1], obs_array["RA"], obs_array["dec"]
            )
            distances = np.atleast_1d(distances)
            # Indexes of observations in obs_array which overlap the DDF
            indxes = np.where(distances <= self.match_radius)[0]

            name_set = set(["ddf_" + name.lower()])
            if indxes.size > 0:
                for indx in indxes:
                    if self.append:
                        prev_values = set(obs_array[self.field_for_label][indx].split(self.separator))
                        new_values = prev_values.union(name_set).difference(self.remove_vals)
                    else:
                        new_values = name_set.difference(self.remove_vals)
                    obs_array[self.field_for_label][indx] = self.separator.join(new_values)
        return obs_array


class LabelRegionsAndDDFs(BaseDetailer):
    """Run both LabelRegionDetailer and LabelDDFDetailer.

    Will always append onto current content of target_name.

    Parameters
    ----------
    label_array : `np.array`
        A HEALpix array of strings that are labels for each HEALpix.
        Default of None uses CurrentAreaMap to load labels.
    field_for_label : `str`
        Which ObservationArray field should be modified.
        Default "target_name".
    camera : `str`
        Which camera model to use to convert a pointing to
        HEALpix IDs. Default "LSST".
    ddf_locations : `dict`
        Dictionary with keys of DDF names and values of RA,dec
        pairs (in degrees). Default of None uses ddf_locations
        utility to get default locations and names.
    match_radius : `float`
        The radius away an observation can be from a DDF center
        and still be tagged. Default 2.0 (degrees)
    separator : `str`
        If append is True, what string to use when appending.
        Default ", ".
    """

    def __init__(
        self,
        label_array=None,
        field_for_label="target_name",
        camera="LSST",
        ddf_location=None,
        match_radius=2.0,
        separator=", ",
    ):
        self.label_detailer = LabelRegionDetailer(
            label_array=label_array,
            field_for_label=field_for_label,
            camera=camera,
            append=True,
            separator=separator,
        )
        self.ddf_detailer = LabelDDFDetailer(
            ddf_location=ddf_location,
            field_for_label=field_for_label,
            append=True,
            match_radius=match_radius,
            separator=separator,
        )

    def __call__(self, obs_array, conditions):

        result = self.label_detailer(obs_array, conditions)
        result = self.ddf_detailer(result, conditions)
        return result


class CopyValueDetailer(BaseDetailer):
    """Copy a value from one observation array column to another

    Parameters
    ----------
    source : `str`
        The name of the source column.
    destination : `str`
        Name of the destination column.
    """

    def __init__(self, source, destination):
        super().__init__()
        self.source = source
        self.destination = destination

    def __call__(self, obs_array, conditions):
        obs_array[self.destination] = obs_array[self.source]
        return obs_array



__all__ = (
    "SolarElongMaskBasisFunction",
    "HaMaskBasisFunction",
    "MoonAvoidanceBasisFunction",
    "MapCloudBasisFunction",
    "PlanetMaskBasisFunction",
    "SolarElongationMaskBasisFunction",
    "AreaCheckMaskBasisFunction",
    "AltAzShadowMaskBasisFunction",
    "AltAzShadowTimeLimitedBasisFunction",
    "RevHaMaskBasisFunction",
    "MaskAllButNES",
    "NInNightMaskBasisFunction",
    "MaskAfterNObsBasisFunction",
    "MaskPoorSeeing",
    "MaskAfterNObsSeeingBasisFunction",
)

import warnings

import healpy as hp
import numpy as np

import rubin_scheduler.scheduler.features as features
from rubin_scheduler.scheduler.basis_functions import BaseBasisFunction
from rubin_scheduler.scheduler.utils import CurrentAreaMap, HpInLsstFov, IntRounded
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD, _angular_separation, _hp_grow_mask


class MaskPoorSeeing(BaseBasisFunction):
    """Mask any part of the sky that is below a seeing limit

    Potential to update to take a declination-dependent seeing limit

    Parameters
    ----------
    bandname : `str`
        The band in which to use the seeing value
    seeing_fwhm_max : `float`
        The seeing FWHM effective to use as a limit (arcsec).
    """

    def __init__(self, bandname, seeing_fwhm_max=1.3, nside=DEFAULT_NSIDE):
        super().__init__(nside=nside)
        self.seeing_fwhm_max = seeing_fwhm_max
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.bandname = bandname

    def __call__(self, conditions, indx=None):
        result = self.result.copy()
        to_mask = np.where(conditions.fwhm_eff[self.bandname] > self.seeing_fwhm_max)
        result[to_mask] = np.nan
        return result


class MaskAfterNObsSeeingBasisFunction(BaseBasisFunction):
    """Mask after a HEALpix has been observed N times. In a
    season.

    Parameters
    ----------
    n_max : `int`
        The maximum number of times to obseve in a season. Default 3.
    bandname : `str`
        The bandname. Default None uses all bands.
    seeing_fwhm_max : `float`
        The seeing limit to use when counting observations (arcsec).
    """

    def __init__(
        self,
        n_max=3,
        nside=DEFAULT_NSIDE,
        bandname=None,
        seeing_fwhm_max=1.3,
        mjd_start=SURVEY_START_MJD,
    ):
        super().__init__(nside=nside)
        self.n_max = n_max
        self.survey_features["nobs"] = features.NObservationsCurrentSeason(
            nside=nside,
            bandname=bandname,
            mjd_start=mjd_start,
            seeing_fwhm_max=seeing_fwhm_max,
        )
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        to_mask = np.where(self.survey_features["nobs"].feature >= self.n_max)[0]
        result[to_mask] = np.nan
        return result


class MaskAllButNES(BaseBasisFunction):
    """Mask everything but the NES region

    Parameters
    ----------
    nes_region : `np.array`
        The index values for HEALpix maps that
        correspond to the NES region. Default None
        uses current default map to label NES.
    """

    def __init__(self, nes_region=None, nside=DEFAULT_NSIDE):
        super().__init__(nside=nside)

        if nes_region is None:
            sag = CurrentAreaMap()
            sky_maps, labels = sag.return_maps()

            nes_region = np.where(labels != "nes")[0]

        self.indx = nes_region
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, **kwargs):
        result = self.result.copy()
        return result


class NInNightMaskBasisFunction(BaseBasisFunction):
    """Mask a HEALpix for the rest of the night
    after it has been observed N times.

    Parameters
    ----------
    n_limit : `int`
        Number of visits to mask after. Default 1.
    bandname : `str`
        Which band to use. Default None assumes all bands.
    """

    def __init__(self, nside=DEFAULT_NSIDE, n_limit=1, bandname=None):

        super(NInNightMaskBasisFunction, self).__init__(nside=nside)
        self.n_limit = n_limit
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.survey_features["n_in_night"] = features.NObsNight(nside=nside, bandname=bandname)

    def _calc_value(self, conditions, indx=None):

        result = self.result.copy()
        to_mask = np.where(self.survey_features["n_in_night"].feature >= self.n_limit)[0]
        result[to_mask] = np.nan
        return result


class RevHaMaskBasisFunction(BaseBasisFunction):
    """Limit the sky based on hour angle

    Parameters
    ----------
    ha_min : float (None)
        The minimum hour angle to accept (hours)
    ha_max : float (None)
        The maximum hour angle to accept (hours)
    """

    def __init__(self, ha_min=None, ha_max=None, nside=DEFAULT_NSIDE):
        super(RevHaMaskBasisFunction, self).__init__(nside=nside)
        self.ha_max = ha_max
        self.ha_min = ha_min
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float) + np.nan

    def _calc_value(self, conditions, **kwargs):
        result = self.result.copy()

        if self.ha_min is not None:
            good = np.where(conditions.HA < (self.ha_min / 12.0 * np.pi))[0]
            result[good] = 0
        if self.ha_max is not None:
            good = np.where(conditions.HA > (self.ha_max / 12.0 * np.pi))[0]
            result[good] = 0

        return result


class MaskAfterNObsBasisFunction(BaseBasisFunction):
    """Mask after a HEALpix has been observed N times.

    Parameters
    ----------
    n_max : `int`
        The maximum number of times. Default 3.
    bandname : `str`
        The bandname. Default None uses all bands.
    """

    def __init__(self, n_max=3, nside=DEFAULT_NSIDE, bandname=None):
        super(MaskAfterNObsBasisFunction, self).__init__(nside=nside)
        self.n_max = n_max
        self.survey_features["nobs"] = features.NObservations(nside=nside, bandname=bandname)
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        to_mask = np.where(self.survey_features["nobs"].feature >= self.n_max)[0]
        result[to_mask] = np.nan
        return result


class SolarElongMaskBasisFunction(BaseBasisFunction):
    """Mask regions larger than some solar elongation limit

    Parameters
    ----------
    elong_limit : float (45)
        The limit beyond which to mask (degrees)
    """

    def __init__(self, elong_limit=45.0, nside=DEFAULT_NSIDE):
        msg = (
            "SolarElongMaskBasisFunction is scheduled for"
            " deprecation, swap to SolarElongationMaskBasisFunction."
        )
        warnings.warn(msg, DeprecationWarning)
        super(SolarElongMaskBasisFunction, self).__init__(nside=nside)
        self.elong_limit = IntRounded(np.radians(elong_limit))
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        to_mask = np.where(IntRounded(conditions.solar_elongation) > self.elong_limit)[0]
        result[to_mask] = np.nan
        return result


class HaMaskBasisFunction(BaseBasisFunction):
    """Limit the sky based on hour angle

    Parameters
    ----------
    ha_min : float (None)
        The minimum hour angle to accept (hours)
    ha_max : float (None)
        The maximum hour angle to accept (hours)
    """

    def __init__(self, ha_min=None, ha_max=None, nside=DEFAULT_NSIDE):
        super(HaMaskBasisFunction, self).__init__(nside=nside)
        self.ha_max = ha_max
        self.ha_min = ha_min
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, **kwargs):
        result = self.result.copy()

        if self.ha_min is not None:
            good = np.where(conditions.HA < (self.ha_min / 12.0 * np.pi))[0]
            result[good] = np.nan
        if self.ha_max is not None:
            good = np.where(conditions.HA > (self.ha_max / 12.0 * np.pi))[0]
            result[good] = np.nan

        return result


class AreaCheckMaskBasisFunction(BaseBasisFunction):
    """Take a list of other mask basis functions, and do an additional
    check for area available"""

    def __init__(self, bf_list, nside=DEFAULT_NSIDE, min_area=1000.0):
        super(AreaCheckMaskBasisFunction, self).__init__(nside=nside)
        self.bf_list = bf_list
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.min_area = min_area

    def check_feasibility(self, conditions):
        result = True
        for bf in self.bf_list:
            if not bf.check_feasibility(conditions):
                return False

        area_map = self.result.copy()
        for bf in self.bf_list:
            area_map *= bf(conditions)

        good_pix = np.where(area_map == 0)[0]
        if hp.nside2pixarea(self.nside, degrees=True) * good_pix.size < self.min_area:
            result = False
        return result

    def _calc_value(self, conditions, **kwargs):
        result = self.result.copy()
        for bf in self.bf_list:
            result *= bf(conditions)
        return result


class SolarElongationMaskBasisFunction(BaseBasisFunction):
    """Mask things at various solar elongations

    Parameters
    ----------
    min_elong : float (0)
        The minimum solar elongation to consider (degrees).
    max_elong : float (60.)
        The maximum solar elongation to consider (degrees).
    """

    def __init__(self, min_elong=0.0, max_elong=60.0, nside=DEFAULT_NSIDE, penalty=np.nan):
        super(SolarElongationMaskBasisFunction, self).__init__(nside=nside)
        self.min_elong = np.radians(min_elong)
        self.max_elong = np.radians(max_elong)
        self.penalty = penalty
        self.result = np.empty(hp.nside2npix(self.nside), dtype=float)
        self.result.fill(self.penalty)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        in_range = np.where(
            (IntRounded(conditions.solar_elongation) >= IntRounded(self.min_elong))
            & (IntRounded(conditions.solar_elongation) <= IntRounded(self.max_elong))
        )[0]
        result[in_range] = 1
        return result


class PlanetMaskBasisFunction(BaseBasisFunction):
    """Mask the bright planets.

    Parameters
    ----------
    mask_radius : float (3.5)
        The radius to mask around a planet (degrees).
    planets : list of str (None)
        A list of planet names to mask. Defaults to ['venus', 'mars',
        'jupiter']. Not including Saturn because it moves really slowly
        and has average apparent mag of ~0.4, so fainter than Vega.

    """

    def __init__(self, mask_radius=3.5, planets=None, nside=DEFAULT_NSIDE, scale=1e5):
        super(PlanetMaskBasisFunction, self).__init__(nside=nside)
        if planets is None:
            planets = ["venus", "mars", "jupiter"]
        self.planets = planets
        self.mask_radius = np.radians(mask_radius)
        self.result = np.zeros(hp.nside2npix(nside))
        # set up a kdtree. Could maybe use healpy.query_disc instead.
        self.in_fov = HpInLsstFov(nside=nside, fov_radius=mask_radius, scale=scale)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        for pn in self.planets:
            indices = self.in_fov(
                np.max(conditions.planet_positions[pn + "_RA"]),
                np.max(conditions.planet_positions[pn + "_dec"]),
            )
            result[indices] = np.nan

        return result


class AltAzShadowMaskBasisFunction(BaseBasisFunction):
    """Mask out a range of altitudes and azimuths, including
     regions which will enter the mask within `shadow_minutes`.

    Masks any alt/az regions as specified by the conditions object in
    `conditions.sky_az_limits` and `conditions.sky_alt_limits`,
    as well as values as provided by `min_alt`, `max_alt`,
    `min_az` and `max_az`.

    Parameters
    ----------
    nside : `int` or None
        HEALpix nside. Default None will look up the package-wide default.
    min_alt : `float`
        Minimum altitude to apply to the mask. Default 20 (degrees).
    max_alt : `float`
        Maximum altitude to allow. Default 82 (degrees).
    min_az : `float`
        Minimum azimuth value to apply to the mask. Default 0 (degrees).
        These azimuth values are absolute azimuth, not cumulative.
        The absolute and cumulative azimuth only diverge if the azimuth
        range is greater than 360 degrees.
    max_az : `float`
        Maximum azimuth value to apply to the mask. Default 360 (degrees).
    shadow_minutes : `float`
        How long to extend masked area in longitude. Default 40 (minutes).
        Choose this value based on the time between when a field might
        be chosen to be scheduled and when it might be observed.
        For pairs, the minimum pair time + some buffer is good.
        For sequences, try the length of the sequence + some buffer.
    pad : `float`
        The value by which to pad the telescope limits, to avoid
        healpix values mapping into pointings from the field tesselations
        which are actually out of bounds. This should typically be
        a bit more than the radius of the fov.  (degrees).
    time_step : `float`
        Calculate mask at multiple timesteps. Useful for when
        shadow minutes is large. Default 10 (minutes).
    """

    def __init__(
        self,
        nside=DEFAULT_NSIDE,
        min_alt=20.0,
        max_alt=86.5,
        min_az=0,
        max_az=360,
        shadow_minutes=40.0,
        pad=3.0,
        scale=1000,
        time_step=10.0,
    ):
        super().__init__(nside=nside)
        self.min_alt = np.radians(min_alt)
        self.max_alt = np.radians(max_alt)
        self.min_az = np.radians(min_az)
        self.max_az = np.radians(max_az)
        self.shadow_time = shadow_minutes / 60.0 / 24.0  # To days
        self.time_step = time_step / 60.0 / 24.0
        self.pad = np.radians(pad)

        self.r_min_alt = IntRounded(self.min_alt)
        self.r_max_alt = IntRounded(self.max_alt)
        self.scale = scale

    def _calc_value(self, conditions, indx=None):
        # Basis function value will be 0 except where masked (then np.nan)
        result = np.zeros(hp.nside2npix(self.nside), dtype=float)

        # For backwards compatiibility:
        try:
            n_steps = np.floor(self.shadow_time / self.time_step)
        except AttributeError:
            n_steps = 0

        if n_steps > 0:
            times = np.arange(n_steps) * self.time_step
            times = times.tolist()
            if self.shadow_time not in times:
                times.append(self.shadow_time)
        else:
            times = [self.shadow_time]

        for time in times:
            # Compute the alt,az values in the future. Use the conditions
            # object so the results are cached and can be used by other
            # surveys is needed. Technically this could fail if the masked
            # region is very narrow or shadow time is very large.
            future_alt, future_az = conditions.future_alt_az(float(np.max(conditions.mjd)) + time)
            r_future_alt = IntRounded(future_alt)
            r_current_alt = IntRounded(conditions.alt)

            # Check the basis function altitude limits, now and future
            result[np.where(r_current_alt < self.r_min_alt)] = np.nan
            result[np.where(r_current_alt > self.r_max_alt)] = np.nan
            result[np.where(r_future_alt < self.r_min_alt)] = np.nan
            result[np.where(r_future_alt > self.r_max_alt)] = np.nan

            # Check the conditions objects 'sky_alt_limit', now and future
            if (conditions.sky_alt_limits is not None) and (len(conditions.sky_alt_limits) > 0):
                combined = np.zeros(hp.nside2npix(self.nside), dtype=float)
                for limits in conditions.sky_alt_limits:
                    # For conditions-based limits, must add pad
                    # And remember that discontinuous areas can be allowed
                    in_bounds = np.ones(hp.nside2npix(self.nside), dtype=float)
                    min_alt = IntRounded(limits[0])
                    max_alt = IntRounded(limits[1])
                    in_bounds[np.where(r_current_alt < min_alt)] = 0
                    in_bounds[np.where(r_current_alt > max_alt)] = 0
                    in_bounds[np.where(r_future_alt < min_alt)] = 0
                    in_bounds[np.where(r_future_alt > max_alt)] = 0
                    combined += in_bounds
                result[np.where(combined == 0)] = np.nan
            # And check against the telescope 'tel_alt_limits'.
            # The tel_alt_limits could be combined with the sky_alt_limits,
            # but it's a bit tricky because sky_alt_limits are (potentially)
            # disjoint allowable areas, while the tel_alt_limits are a single
            # wide set of allowable area which explicitly disallows anything
            # outside that range. The az limits versions are similar.
            if conditions.tel_alt_limits is not None:
                min_alt = IntRounded(conditions.tel_alt_limits[0])
                max_alt = IntRounded(conditions.tel_alt_limits[1])
                result[np.where(r_current_alt < min_alt)] = np.nan
                result[np.where(r_current_alt > max_alt)] = np.nan
                result[np.where(r_future_alt < min_alt)] = np.nan
                result[np.where(r_future_alt > max_alt)] = np.nan

            # note that % (mod) is not supported for IntRounded
            two_pi = 2 * np.pi
            # Check the basis function azimuth limits, now and future
            if np.abs(self.max_az - self.min_az) < two_pi:
                az_range = (self.max_az - self.min_az) % (two_pi)
                out_of_bounds = np.where((conditions.az - self.min_az) % (two_pi) > az_range)[0]
                result[out_of_bounds] = np.nan
                out_of_bounds = np.where((future_az - self.min_az) % (two_pi) > az_range)[0]
                result[out_of_bounds] = np.nan
            # Check the conditions objects azimuth limits, now and future
            if (conditions.sky_az_limits is not None) and (len(conditions.sky_az_limits) > 0):
                combined = np.zeros(hp.nside2npix(self.nside), dtype=float)
                for limits in conditions.sky_az_limits:
                    in_bounds = np.ones(hp.nside2npix(self.nside), dtype=float)
                    min_az = limits[0]
                    max_az = limits[1]
                    if np.abs(max_az - min_az) < two_pi:
                        az_range = (max_az - min_az) % (two_pi)
                        out_of_bounds = np.where((conditions.az - min_az) % (two_pi) > az_range)[0]
                        in_bounds[out_of_bounds] = 0
                        out_of_bounds = np.where((future_az - min_az) % (two_pi) > az_range)[0]
                        in_bounds[out_of_bounds] = 0
                    combined += in_bounds
                result[np.where(combined == 0)] = np.nan
            # Check against the kinematic hard limits.
            if conditions.tel_az_limits is not None:
                if np.abs(conditions.tel_az_limits[1] - conditions.tel_az_limits[0]) < two_pi:
                    az_range = (conditions.tel_az_limits[1] - conditions.tel_az_limits[0]) % (two_pi)
                    out_of_bounds = np.where(
                        (conditions.az - conditions.tel_az_limits[0]) % (two_pi) > az_range
                    )[0]
                    result[out_of_bounds] = np.nan
                    out_of_bounds = np.where((future_az - conditions.tel_az_limits[0]) % (two_pi) > az_range)[
                        0
                    ]
                    result[out_of_bounds] = np.nan

        # Grow the resulting mask by self.pad, to avoid field centers
        # falling outside the boundaries of what is actually reachable.
        if self.pad > 0:
            mask_indx = np.where(np.isnan(result))[0]
            to_mask_indx = _hp_grow_mask(self.nside, tuple(mask_indx), scale=self.scale, grow_size=self.pad)
            result[to_mask_indx] = np.nan

        return result


class AltAzShadowTimeLimitedBasisFunction(AltAzShadowMaskBasisFunction):
    """Similar to AltAzShadowMaskBasisFunction, but only adds a mask
    if the time within the night is close to the specified sunrise/sunset
    event

    Parameters
    ----------
    <all the kwargs from AltAzShadowMaskBasisFunction>

    time_to_sun : `float`
        How close in time one must be to the sun up/down time (hours).
        Default 2.
    sun_keys : `list`
        List of strings that are looked up in the conditions object.
        Default ["sunrise", "sunset"]. Options currently are
        sunrise, sunset, sun_n12_setting, sun_n18_setting,
        sun_n18_rising, sun_n12_rising
    """

    def __init__(self, time_to_sun=2.0, sun_keys=["sunrise", "sunset"], **kwargs):
        self.time_to_sun = time_to_sun / 24.0  # to days
        self.sun_keys = sun_keys
        super().__init__(**kwargs)

    def _calc_value(self, conditions, indx=None):

        time_to_sun = [conditions.mjd - getattr(conditions, key) for key in self.sun_keys]
        time_to_sun = np.abs(time_to_sun)
        # We are close to a sunrise/set event, apply mask
        if np.min(time_to_sun) < self.time_to_sun:
            return super()._calc_value(conditions, indx=indx)
        else:
            # In the middle of the night, no mask
            return 0


class MoonAvoidanceBasisFunction(BaseBasisFunction):
    """Avoid observing within `moon_distance` of the moon.

    Parameters
    ----------
    moon_distance: float (30.)
        Minimum allowed moon distance. (degrees)

    Notes
    -----
    As the current specified requirements for the observatory
    are "observe more than 30 degrees from the moon", this basis
    function simply does that at present. This includes
    times the moon is below the horizon or if the moon close to new.
    Most likely, this avoidance region should depend on lunar phase,
    the band used for observations, and whether the moon is above
    or below the horizon.
    """

    def __init__(self, nside=DEFAULT_NSIDE, moon_distance=30.0):
        super(MoonAvoidanceBasisFunction, self).__init__(nside=nside)
        self.update_on_newobs = False

        self.moon_distance = IntRounded(np.radians(moon_distance))
        self.result = np.ones(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()

        angular_distance = _angular_separation(
            conditions.az, conditions.alt, conditions.moon_az, conditions.moon_alt
        )

        result[IntRounded(angular_distance) < self.moon_distance] = np.nan

        return result


class BulkCloudBasisFunction(BaseBasisFunction):
    """Mark healpixels on a map if their cloud values are greater than
    the same healpixels on a maximum cloud map.

    Parameters
    ----------
    nside: int (default_nside)
        The healpix resolution.
    max_cloud_map : numpy array (None)
        A healpix map showing the maximum allowed cloud values for all
        points on the sky
    out_of_bounds_val : float (10.)
        Point value to give regions where there are no observations
        requested
    """

    def __init__(self, nside=DEFAULT_NSIDE, max_cloud_map=None, max_val=0.7, out_of_bounds_val=np.nan):
        super(BulkCloudBasisFunction, self).__init__(nside=nside)
        self.update_on_newobs = False

        if max_cloud_map is None:
            self.max_cloud_map = np.zeros(hp.nside2npix(nside), dtype=float) + max_val
        else:
            self.max_cloud_map = max_cloud_map
        self.out_of_bounds_area = np.where(self.max_cloud_map > 1.0)[0]
        self.out_of_bounds_val = out_of_bounds_val
        self.result = np.ones(hp.nside2npix(self.nside))

    def _calc_value(self, conditions, indx=None):
        """
        Parameters
        ----------
        indx : list (None)
            Index values to compute, if None, full map is computed
        Returns
        -------
        Healpix map where pixels with a cloud value greater than the
        max_cloud_map value are marked as unseen.
        """

        result = self.result.copy()

        clouded = np.where(self.max_cloud_map <= conditions.bulk_cloud)
        result[clouded] = self.out_of_bounds_val

        return result


class MapCloudBasisFunction(BaseBasisFunction):
    """Mark healpixels on a map if their cloud values are greater than
    the same healpixels on a maximum cloud map. Currently a placeholder for
    when the telemetry stream can include a full sky cloud map.

    Parameters
    ----------
    nside: int (default_nside)
        The healpix resolution.
    max_cloud_map : numpy array (None)
        A healpix map showing the maximum allowed cloud values for all
        points on the sky
    out_of_bounds_val : float (10.)
        Point value to give regions where there are no observations
        requested
    """

    def __init__(self, nside=DEFAULT_NSIDE, max_cloud_map=None, max_val=0.7, out_of_bounds_val=np.nan):
        super().__init__(nside=nside)
        self.update_on_newobs = False

        if max_cloud_map is None:
            self.max_cloud_map = np.zeros(hp.nside2npix(nside), dtype=float) + max_val
        else:
            self.max_cloud_map = max_cloud_map
        self.out_of_bounds_area = np.where(self.max_cloud_map > 1.0)[0]
        self.out_of_bounds_val = out_of_bounds_val
        self.result = np.ones(hp.nside2npix(self.nside))

    def _calc_value(self, conditions, indx=None):
        """
        Parameters
        ----------
        indx : list (None)
            Index values to compute, if None, full map is computed
        Returns
        -------
        Healpix map where pixels with a cloud value greater than the
        max_cloud_map value are marked as unseen.
        """

        result = self.result.copy()

        clouded = np.where(self.max_cloud_map <= conditions.bulk_cloud)
        result[clouded] = self.out_of_bounds_val

        return result



import os
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.utils import SchemaConverter
from rubin_scheduler.utils.consdb import ConsDBVisits, load_consdb_visits


class TestConsdb(unittest.TestCase):

    @unittest.skip("avoid requiring access to consdb for tests.")
    def test_consdb_read_visits_lsstcam(self):
        day_obs: str = "2025-07-01"
        consdb_visits: ConsDBVisits = load_consdb_visits("lsstcam", day_obs)
        schema_converter: SchemaConverter = SchemaConverter()

        opsim: pd.DataFrame = consdb_visits.opsim
        schema_converter.opsimdf2obs(opsim)

        obs: np.recarray = consdb_visits.obs
        schema_converter.obs2opsim(obs)

    @unittest.skip("avoid requiring access to consdb for tests.")
    def test_consdb_read_visits_with_constraints(self):
        day_obs: str = "2025-07-01"
        constraints = "band = 'z'"
        consdb_visits: ConsDBVisits = load_consdb_visits("lsstcam", day_obs, constraints=constraints)
        opsim: pd.DataFrame = consdb_visits.opsim
        assert len(opsim) > 0
        assert np.all(opsim.band == "z")

    @unittest.skip("avoid requiring access to consdb for tests.")
    def test_consdb_read_visits_comcamsim(self):
        day_obs: str = "2024-06-26"
        consdb_visits: ConsDBVisits = load_consdb_visits("lsstcomcamsim", day_obs)
        schema_converter: SchemaConverter = SchemaConverter()

        opsim: pd.DataFrame = consdb_visits.opsim
        schema_converter.opsimdf2obs(opsim)

        obs: np.recarray = consdb_visits.obs
        schema_converter.obs2opsim(obs)

    @unittest.skip("avoid requiring access to consdb for tests.")
    def test_consdb_read_visits_comcam(self):
        day_obs: str = "2024-12-10"
        consdb_visits: ConsDBVisits = load_consdb_visits("lsstcomcam", day_obs)
        schema_converter: SchemaConverter = SchemaConverter()

        opsim: pd.DataFrame = consdb_visits.opsim
        schema_converter.opsimdf2obs(opsim)

        obs: np.recarray = consdb_visits.obs
        schema_converter.obs2opsim(obs)

    @unittest.skip("avoid requiring access to consdb for tests.")
    def test_consdb_read_visits_latiss(self):
        day_obs: str = "2025-03-03"
        consdb_visits: ConsDBVisits = load_consdb_visits("latiss", day_obs)
        schema_converter: SchemaConverter = SchemaConverter()

        opsim: pd.DataFrame = consdb_visits.opsim
        schema_converter.opsimdf2obs(opsim)

        obs: np.recarray = consdb_visits.obs
        schema_converter.obs2opsim(obs)

    @unittest.skip("avoid requiring access to consdb for tests.")
    def test_consdb_read_visits_with_token_file(self):
        if "ACCESS_TOKEN" in os.environ:
            token = os.environ["ACCESS_TOKEN"]
        else:
            try:
                import lsst.rsp

                token = lsst.rsp.get_access_token()
            except ImportError:
                return

        with TemporaryDirectory() as temp_dir:
            token_file = Path(temp_dir).joinpath("test_file")
            with open(token_file, "w") as token_io:
                token_io.write(token)

            day_obs: str = "2025-06-20"
            consdb_visits: ConsDBVisits = load_consdb_visits("lsstcam", day_obs, token_file=token_file)
            schema_converter: SchemaConverter = SchemaConverter()

            opsim: pd.DataFrame = consdb_visits.opsim
            schema_converter.opsimdf2obs(opsim)

            obs: np.recarray = consdb_visits.obs
            schema_converter.obs2opsim(obs)



import os
import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.model_observatory import ModelObservatory

SAMPLE_BIG_DATA_FILE = os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")


class TestExample(unittest.TestCase):
    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_example(self):
        """Try out the example scheduler."""
        mjd_start = utils.SURVEY_START_MJD
        nside = 32
        survey_length = 4.0  # days
        scheduler = example_scheduler(nside=nside, mjd_start=mjd_start)
        observatory = ModelObservatory(
            nside=nside, mjd_start=mjd_start, cloud_data="ideal", downtimes="ideal"
        )
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None
        )
        # Check that greedy observed some
        assert "greedy" in observations["scheduler_note"]
        # check some long pairs got observed
        assert np.any(["pair_33" in obs for obs in observations["scheduler_note"]])
        # Make sure lots of observations executed
        assert observations.size > 1000
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0
        # Make sure a twilight executed
        assert np.any(["twilight_near_sun, 1" in note for note in observations["scheduler_note"]])


if __name__ == "__main__":
    unittest.main()



__all__ = ("CloudMap",)

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.utils import match_hp_resolution
from rubin_scheduler.utils import DEFAULT_NSIDE


class CloudMap(object):
    """A class to hold cloud maps over time

    Parameters
    ----------
    nside_out : `int`
        The nside to output maps as. Default to
        DEFAULT_NSIDE (probably 32).
    time_limit : `float`
        Do not return a cloud map if there is nothing within
        the time limit. Default 20 (minutes).
    """

    def __init__(self, nside_out=DEFAULT_NSIDE, time_limit=20.0, max_frames=2000):
        self.time_limit = time_limit / 60 / 24.0  # to days
        self.nside_out = nside_out
        self.mjds = []
        self.cloud_extinction_hparrays = []
        self.cloud_extinction_uncerts = []
        self.max_frames = max_frames

    def add_frame(self, input_cloud_extinction, mjd, nested=False, uncert=None):
        """Add a frame. Will be converted to ring order if needed
        and set to the self.nside_out resolution.

        Parameters
        ----------
        input_cloud_extinction : `np.array`
            HEALpix array with extinction in mags.
        mjd : `float`
           The MJD of the cloud frame
        nested : `bool`
            If True, converts the incoming map to ring order.
            Default False
        uncert : `np.array`
            HEALpix array with uncertainty in extinction magnitudes.
        """
        self.mjds.append(mjd)

        to_add = input_cloud_extinction.copy()
        if uncert is None:
            to_add_uncert = to_add * 0
        else:
            to_add_uncert = uncert.copy()

        if nested:
            to_add = hp.reorder(to_add, n2r=True)
            to_add_uncert = hp.reorder(to_add_uncert, n2r=True)

        # Set to the proper resolution.
        to_add = match_hp_resolution(to_add, self.nside_out)
        to_add_uncert = match_hp_resolution(to_add_uncert, self.nside_out)

        self.cloud_extinction_hparrays.append(to_add)
        self.cloud_extinction_uncerts.append(to_add_uncert)

        # If things were entered out of order
        if not np.all(self.mjds[:-1] <= self.mjds[1:]):
            order = np.argsort(self.mjds)
            self.mjds = self.mjds[order]
            self.cloud_extinction_hparrays = self.cloud_extinction_hparrays[order]
            self.cloud_extinction_uncerts = self.cloud_extinction_uncerts[order]

        # If we are getting to the limit of out size
        while len(self.mjds) > self.max_frames:
            del self.cloud_extinction_hparrays[0]
            del self.cloud_extinction_uncerts[0]
            del self.mjds[0]

    def extinction_closest(self, mjd, hpid=None, uncert=False):
        """Return the closest map

        Parameters
        ----------
        mjd : `float`
            MJD of the desired extinction map.
        hpid : `int`
            If only a subset of the HEALpix array is needed, can
            specify index.
        uncert : `bool`
            Also return the uncertainty map with the extinction map.

        Returns
        -------
        Healpix array in ring order with cloud extinction in magnitudes.

        """
        diff = np.abs(np.array(self.mjds) - mjd)
        # If we don't have a close enough map, just return zero
        if np.min(diff) > self.time_limit:
            if uncert:
                return 0, 0
            else:
                return 0
        indx = np.min(np.where(diff == diff.min())[0])
        if hpid is not None:
            if uncert:
                return self.cloud_extinction_hparrays[indx][hpid], self.cloud_extinction_uncerts[indx][hpid]
            else:
                return self.cloud_extinction_hparrays[indx][hpid]

        if uncert:
            return self.cloud_extinction_hparrays[indx], self.cloud_extinction_uncerts[indx]
        else:
            return self.cloud_extinction_hparrays[indx]

    def extinction_forecast(self, mjd):
        """Spot to put in projected cloud extinction in the future"""
        raise NotImplementedError("Not done yet")



import unittest

import numpy as np

from rubin_scheduler.utils.code_utilities import _validate_inputs


class CodeUtilsTest(unittest.TestCase):
    def test_validate_inputs(self):
        """
        test that _validate_inputs returns expected values
        """

        # test that it correctly identifies numpy array inputs
        self.assertTrue(_validate_inputs([np.array([1, 2]), np.array([3, 4])], ["a", "b"], "dummy_method"))

        # test that it correctly identifies inputs that are numbers
        self.assertFalse(_validate_inputs([1, 2, 3], ["a", "b", "c"], "dummy"))

        # test that the correct exception is raised when you pass in
        # a number a numpy array
        with self.assertRaises(RuntimeError) as ee:
            _validate_inputs([1, np.array([2, 3])], ["a", "b"], "dummy")

        self.assertIn("and the same type", ee.exception.args[0])

        # test that the correct exception is raised when you pass in
        # numpy arrays of different length
        with self.assertRaises(RuntimeError) as ee:
            _validate_inputs([np.array([1, 2]), np.array([1, 2, 3])], ["a", "b"], "dummy")

        self.assertIn("same length", ee.exception.args[0])

        # test that an exception is raised if lists (rather than numpy
        # arrays) are passed in
        with self.assertRaises(RuntimeError) as ee:
            _validate_inputs([[1, 2], [3, 4]], ["a", "b"], "dummy")

        self.assertIn("either a number or a numpy array", ee.exception.args[0])


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from rubin_scheduler.utils import SharedNumpyArray


class ParallelTools(unittest.TestCase):
    def test_shared_array(self):
        input_array = np.arange(100).reshape(10, 10)

        sharred_array = SharedNumpyArray(input_array)

        assert np.array_equal(input_array, sharred_array.read())
        assert np.array_equal(input_array, sharred_array.copy())

        sharred_array.unlink()


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import simple_greedy_survey
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler

SAMPLE_BIG_DATA_FILE = os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")


class ModelObservatoryWindy(ModelObservatory):
    """Have the model observatory always have a strong wind from the north"""

    def return_conditions(self):
        """
        Returns
        -------
        rubin_scheduler.scheduler.features.conditions object
        """
        self.conditions = super().return_conditions()

        # Always have a strong wind from the north
        wind_speed = 40.0
        wind_direction = 0.0
        self.conditions.wind_speed = wind_speed
        self.conditions.wind_direction = wind_direction

        return self.conditions


class TestWind(unittest.TestCase):
    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_wind(self):
        """
        Test that a wind mask prevent things from being executed in
        the wrong spot
        """
        mjd_start = utils.SURVEY_START_MJD
        nside = 32
        survey_length = 1.0  # days

        surveys = [simple_greedy_survey(bandname=f) for f in "gri"]

        scheduler = CoreScheduler(surveys, nside=nside)
        observatory = ModelObservatoryWindy(
            nside=nside, mjd_start=mjd_start, downtimes="ideal", cloud_data="ideal"
        )
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None, n_visit_limit=501
        )

        # Make sure lots of observations executed, but allow short night
        # if survey_start changes
        assert observations.size > 500
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0

        # Make sure nothing executed in the strong wind
        assert np.min(np.degrees(observations["az"])) > 30.0
        assert np.max(np.degrees(observations["az"])) < (360.0 - 30.0)


if __name__ == "__main__":
    unittest.main()



import unittest

import healpy as hp
from astropy.coordinates import AltAz, EarthLocation, get_sun
from astropy.time import Time

import rubin_scheduler.utils as utils


class Constantsests(unittest.TestCase):

    def test_survey_start(self):
        ss_mjd = utils.SURVEY_START_MJD
        ss_time = Time(ss_mjd, format="mjd")

        # Demand that the default start date be during the day
        site = utils.Site("LSST")
        location = EarthLocation(lat=site.latitude, lon=site.longitude, height=site.height)
        sun = get_sun(ss_time)
        aa = AltAz(location=location, obstime=ss_time)
        sun_aa = sun.transform_to(aa)
        assert sun_aa.alt > 0

    def test_nside(self):
        nside = utils.DEFAULT_NSIDE
        npix = hp.nside2npix(nside)
        assert npix > 0


if __name__ == "__main__":
    unittest.main()
